;;; $Id: define-datatype.tst,v 1.1 2005/12/22 22:55:24 leavens Exp $

(load "test-harness.scm")
(load "test-suite.scm")

;;; The bottom part of define-datatype.scm from the original version.
;;; ------------------------------
;;; general helpers

(define always?
  (lambda (x) #t))

(define list-of
  (lambda (pred . l)
    (let ((all-preds (cons pred l)))
      (lambda (obj)
        (let loop ((obj obj) (preds '()))
          (or 
            ;; if list is empty, preds should be, too
            (and (null? obj) (null? preds))
            (if (null? preds)
                ;; if preds is empty, but list isn't, then recycle
                (loop obj all-preds)
                ;; otherwise check and element and recur.
                (and (pair? obj)
                     ((car preds) (car obj))
                     (loop (cdr obj) (cdr preds))))))))))

;;; ------------------------------
;;; examples

(define-datatype btree btree?
  (empty-btree)
  (btree-node (left btree?) (key integer?) (right btree?)))

(define sort-intlist
  (letrec ((flatten-btree
             (lambda (bt acc)
               (cases btree bt
                 (empty-btree () acc)
                 (btree-node (left key right)
                   (flatten-btree left
                     (cons key
                       (flatten-btree right acc)))))))
           (insert-list
             (lambda (ls bt)
               (if (null? ls) bt
                   (insert-list (cdr ls) (insert (car ls) bt)))))
           (insert
             (lambda (n bt)
               (cases btree bt
                 (empty-btree ()
                   (btree-node (empty-btree) n (empty-btree)))
                 (btree-node (left key right)
                   (cond
                     ((equal? n key) bt)
                     ((< n key)
                      (btree-node (insert n left) key right))
                     (else
                       (btree-node left key (insert n right)))))))))
    (lambda (ls)
      (flatten-btree (insert-list ls (empty-btree)) '()))))

(define define-datatype:test0
  '(sort-intlist '(8 6 7 5 3 0 9)))

;;; ------------------------------

(define-datatype lyst lyst?
  (nil)
  (pair
    (head always?)
    (tail lyst?)))

(define list->lyst
  (lambda (ls)
    (if (null? ls)
        (nil)
        (pair (car ls) (list->lyst (cdr ls))))))

(define lyst->list                      ; this tests hygiene
  (lambda (pr)
    (cases lyst pr
      (nil () '())
      (pair (head tail)
        (cons head (lyst->list tail))))))

(define define-datatype:test1
  '(lyst->list (list->lyst '(this is a weird form of identity))))

(define lyst-nil?                       ; this tests else-ability
  (lambda (pr)
    (cases lyst pr
      (nil () #t)
      (else #f))))

(define define-datatype:test2
  '(list (lyst-nil? (nil)) (lyst-nil? (pair 3 (nil)))))

(define define-datatype:test3
  '(begin
     (define-datatype alist alist?
       (anil)
       (apair (head always?) (tail blist?)))
     (define-datatype blist blist?
       (bnil)
       (bpair (head always?) (tail alist?)))
     (apair 5 (bpair 4 (anil)))))

(define define-datatype:test4
  '(begin
     (define-datatype fff fff?
       (wu)
       (bu (fff fff?)))
     (let ((fff 3))
       (fff? (bu (wu))))))

;;; ------------------------------
;;; error tests

(define define-datatype:err0            ; wrong # args to a constructor
  '(pair))

(define define-datatype:err1            ; wrong type args to a constructor
  '(pair 3 4))

(define define-datatype:err2            ; unlisted variant in cases
  '(cases lyst (nil)
     (nil () 3)))

(define define-datatype:err3            ; wrong type argument to case
  '(cases lyst 5
     (pair (x y) 3)
     (nil () 8)))

(define define-datatype:err4            ; wrong # fields in cases
  '(cases lyst (nil)
     (pair (y z) 4)
     (nil (x) 3)
     (else 5)))

(define define-datatype:err5            ; misspelled variant in cases
  '(cases lyst (nil)
     (ppair (y z) 4)
     (nil () 8)
     (else 5)))

(define define-datatype:err10           ; non-symbol used for variant name
  '(define-datatype x x?
     ((r) (a b))))

(define define-datatype:err11           ; duplicate variant names
  '(define-datatype x x?
     (r (zoo goo?))
     (r (foo goo?))
     (s (joo goo?))))

(define define-datatype:err14           ; only type name
  '(define-datatype a a?))

(define define-datatype:err18           ; duplicate variant clauses
  '(cases lyst (nil)
     (nil () 3)
     (nil () 4)))

(define define-datatype:err19           ; repeated variant name.
  '(begin
     (define-datatype www www?
       (foo (man www?)))
     (define-datatype zzz zzz?
       (foo (harry zzz?)))))

;; uuu? is undefined.  Should it be uu?
(define define-datatype:err20           ; isa's symbol arg is not a type name
  '(begin
     (define-datatype uu uu?
       (goo (man number?)))
     (define-datatype:pretty-print (uu? (goo 5)))
     (uuu? (goo 6))))

(define define-datatype:err21           ; Too many args to uuuu?
  '(begin
     (define-datatype uuuu uuuu?
       (gu)
       (zu (foo uuuu?)))
     (uuuu? (zu (zu (gu))) 5)))

;; what is the "?" on the last line??
(define define-datatype:err22           ; Too few args to isa
  '(begin
     (define-datatype uuuu uuuu?
       (gu)
       (zu (foo uuuu?)))
     ( (zu (zu (gu)))?)))

(define define-datatype:err23           ; Too many args to isa
  '(begin
     (define-datatype uuuu uuuu?
       (gu)
       (zu (foo uuuu?)))
     (uuuu? (zu (zu (gu))))))

(define define-datatype:err24           ; type name cannot be chosen
  '(begin                               ; from existing variant name
     (define-datatype uuuu uuuu?
       (gu)
       (zu (foo uuuu?)))
     (define-datatype gu gu?
       (hood))
     (uuuu? (zu (zu (gu))))))

(define define-datatype:err25           ; type name and constructor name
  '(begin                               ; cannot be the same.
     (define-datatype uuuu uuuu?
       (gu)
       (uuuu (foo uuuu?)))
     (uuuu? (zu (zu (gu))))))

(define define-datatype:err26           ; variantr name cannot be chosen
  '(begin                               ; from existing type name
     (define-datatype uuuu uuuu?
       (gu)
       (uuuu^ (foo uuuu?)))
     (define-datatype gru gru?
       (uuuu))
     (uuuu? (zu (zu (gu))))))



;; wdc: what is this 5 ?
(define define-datatype:err27           ; isa's arg is not a symbol.
  '(begin                               
     (define-datatype uuuu uuuu?
       (gu)
       (uuuu^ (foo uuuu?)))
     (5 (zu (zu (gu))))))



(define define-datatype:err28           ; 1st & 2nd arg of cases should not
  '(begin                               ; be the same.
     (define-datatype uuuu** uuuu**?
       (gud)
       (uuuu^^ (foo uuuu**?)))
     (let ((uuuu** (uuuu^^ (uuuu^^ (gud)))))
       (cases uuuu** uuuu**
         (gud () "Hello")
         (uuuu^^ (foo) "Goodbye")))))

(define define-datatype:err29           ; 2nd arg of cases should not
  '(begin                               ; be the same as any variant name.
     (define-datatype uuuu** uuuu**?
       (gud)
       (uuuu^^ (foo uuuu**?)))
     (let ((uuuu^^ (uuuu^^ (uuuu^^ (gud)))))
       (cases uuuu** uuuu^^
         (gud () "Hello")
         (uuuu^^ (foo) "Goodbye")))))

(define do-all-tests
  (let ((tests (list
                 define-datatype:test0
                 define-datatype:test1
                 define-datatype:test2
                 define-datatype:test3
                 define-datatype:test4  
                 define-datatype:err0
                 define-datatype:err1
                 define-datatype:err2
                 define-datatype:err3
                 define-datatype:err4
                 define-datatype:err5
                 define-datatype:err10
                 define-datatype:err11 
                 define-datatype:err14 
                 define-datatype:err18
                 define-datatype:err19
                 define-datatype:err21
                 define-datatype:err23 
                 define-datatype:err24
                 define-datatype:err25
                 define-datatype:err26
                 define-datatype:err28
                 define-datatype:err29))) 
    (lambda (chezer)
      (for-each chezer tests))))

; mw added dynamic-wind around rebinding of eopl:error-stop
(define define-datatype:tester
  (lambda (example)
    (display "------------------------------")
    (newline)
    (sllgen:pretty-print example)
    (display "-->")
    (newline)
    (call-with-current-continuation
     (lambda (k)
       (let ((alpha (lambda () (k #f))))
         (let ((swap (lambda ()
                     (let ((temp eopl:error-stop))
                       (set! eopl:error-stop alpha)
                       (set! alpha temp)))))
           (dynamic-wind
             swap
             (lambda () 
               (write (eval example (interaction-environment)))
               (newline)
               #t)
             swap)))))))

(define define-datatype:test-all
  (lambda ()
    (do-all-tests define-datatype:tester)
    (define-datatype:reset-registries)))
