;;; $Id: tc-parsing-stream.scm,v 1.5 2006/02/06 01:09:10 leavens Exp $
;;; Copyright (C) 2006 Iowa State University
;;;
;;; This file is part of Typedscm.
;;;
;;; This library is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU Lesser General Public License
;;; as published by the Free Software Foundation; either version 2.1,
;;; of the License, or (at your option) any later version.
;;;
;;; This library is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; Lesser General Public License for more details.
;;;
;;; You should have received a copy of the GNU Lesser General Public
;;; License along with Typedscm; see the file LesserGPL.txt.  If not,
;;; write to the Free Software Foundation, Inc., 51 Franklin St, Fifth
;;; Floor, Boston, MA 02110-1301 USA.

;;; streams for parsing

(module tc-parsing-stream (lib "typedscm.ss" "typedscm")

(provide tc:stream-of tc:stream-car tc:stream-cdr tc:make-stream
         tc:the-null-stream tc:stream-null? tc:stream-for-each
         tc:make-input-stream) 
  
;;; EOPL Figure F.1 : page 493

(deftype tc:stream-of
  (forall (T) (-> ((type-predicate-for T))
		  (type-predicate-for (tc:stream-of T)))))
(deftype tc:stream-car (forall (T) (-> ((tc:stream-of T)) T)))
(deftype tc:stream-cdr (forall (T) (-> ((tc:stream-of T)) (tc:stream-of T))))
(deftype tc:make-stream
  (forall (T) (-> (T (-> () (tc:stream-of T))) (tc:stream-of T))))
(deftype tc:the-null-stream (forall (T) (-> () (tc:stream-of T))))
(deftype tc:stream-null? (forall (T) (-> ((tc:stream-of T)) boolean)))
(deftype tc:stream-for-each
  (forall (T) (-> ((-> (T) void) (tc:stream-of T)) void)))
(deftype tc:make-input-stream (-> () (tc:stream-of char)))

(require (lib "relax-type-checking.scm" "typedscm"))

(defrep
  (forall (T) (tc:stream-of T) (-> () (pair-of T (-> () (tc:stream-of T))))))

(define tc:stream-of
  (has-type-trusted
    (forall (T) (-> ((type-predicate-for T))
		    (type-predicate-for (tc:stream-of T))))
    (forall (T?)
      (lambda (x)
	(procedure? x)))))

(define tc:stream-car
  (lambda (s)
    (pair-first (s))))

(define tc:stream-cdr
  (has-type-trusted
    (forall (T) (-> ((-> () (pair-of T (-> () (tc:stream-of T)))))
		    (-> () (pair-of T (-> () (tc:stream-of T))))))
    (lambda (s)
      (cdr (s)))))

(define tc:make-stream
  (has-type-trusted
    (forall (T) (-> (T (-> () (-> () (pair-of T (-> () (tc:stream-of T))))))
		    (-> () (pair-of T (-> () (tc:stream-of T))))))
    (lambda (car-val th)
      (lambda ()
	(make-pair
	 car-val
	 (let ((cdr-val #f))
	   (lambda ()
	     (if (pair? cdr-val)
		 cdr-val
		 (begin (set! cdr-val ((th)))
			cdr-val)))))))))

(define tc:the-null-stream
  (has-type-trusted
    (forall (T) (-> () (pair-of T (-> () (tc:stream-of T)))))
    (lambda ()
      (tc:make-stream 'end-of-stream tc:the-null-stream))))

(define tc:stream-null?
  (lambda (s)
    (eq? (tc:stream-car s) 'end-of-stream)))

;;; Figure F.2 : page 494

(define tc:stream-for-each
  (lambda (proc stream)
    (letrec
      ((loop
         (lambda (stream)
           (if (not (tc:stream-null? stream))
               (begin
                 (proc (tc:stream-car stream))
                 (loop (tc:stream-cdr stream)))))))
      (loop stream))))

(define tc:make-input-stream
  (lambda ()
    (let ((char (read-char)))
      (if (eof-object? char)
          (tc:the-null-stream)
	    (tc:make-stream char tc:make-input-stream)))))

) ;; end module
