;;; $Id: tc-scheme-parser.tst,v 1.12 2006/02/10 05:23:31 leavens Exp $
;;; Copyright (C) 2006 Iowa State University
;;;
;;; This file is part of Typedscm.
;;;
;;; This library is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU Lesser General Public License
;;; as published by the Free Software Foundation; either version 2.1,
;;; of the License, or (at your option) any later version.
;;;
;;; This library is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; Lesser General Public License for more details.
;;;
;;; You should have received a copy of the GNU Lesser General Public
;;; License along with Typedscm; see the file LesserGPL.txt.  If not,
;;; write to the Free Software Foundation, Inc., 51 Franklin St, Fifth
;;; Floor, Boston, MA 02110-1301 USA.

(require (lib "tc-scheme-parser.scm" "typedscm")
	 (lib "tc-scheme-unparser.scm" "typedscm"))

;; The first batch here is for hand debugging of the parser
;; Uncomment various bits to see details.

; (pretty-print
;  (tc:character-string-exp-parser
;   "1"))
; 
; (pretty-print
;  (tc:character-string-form-parser
;   "1"))
; 
; (pretty-print
;  (tc:character-string-form-parser
;   "(define a 1)"))
; 
; (pretty-print
;  (tc:character-string-form-parser
;   "(define (foo x) 1)"))
; 
; (pretty-print
;  (tc:character-string-form-parser
;   "(define (bar y . z) 2)"))
; 
; (pretty-print
;  (tc:character-string-form-parser
;   "(define (foo x)
;      (define a 1)
;      (define b 2)
;      3
;      4)"))
; 
; (pretty-print
;  (tc:character-string-form-parser
;   "(define (foo x)
;      (define a 1)
;      (define (b z q r . m)
;        (define c 3)
;        5 6 2)
;      0
;      3
;      4)"))
; 
; ;; diff between grouped defines and begin forms in the next two
; (pretty-print
;   (tc:character-string-program-parser
;    "(begin (deftype a number) (define a 1) (define b 2))"))
; 
; (pretty-print
;   (tc:character-string-program-parser
;    "(begin (deftype a number) (define a 1) (set! a 3) (define b 2))"))

(run-regression-tests-equal?
 '(

   ( (tc:unparse-scheme-exp
      (tc:character-string-exp-parser
       "1"))
     ==> 1 )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "(define a 1)"))
     ==> (define a 1) )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "(define (foo x) 1)"))
     ==> (define (foo x) 1) )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "(define (bar y . z) 2)"))
     ==> (define (bar y . z) 2) )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "(define (foo x)
          (define a 1)
          (define b 2)
          3
          4)"))
     ==> (define (foo x)
	   (define a 1)
	   (define b 2)
	   (begin 3 4)) )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "(define (foo x)
          (define a 1)
          (define (b z q r . m) (define c 3)  5 6 2)
          0 3 4)"))
     ==> (define (foo x)
	   (define a 1)
	   (define (b z q r . m) (define c 3) (begin 5 6 2))
	   (begin 0 3 4)) )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "#t"))
     ==> #t )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "\"a string\""))
     ==> "a string" )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "#\\c"))
     ==> #\c )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "x-is-a-varref"))
     ==> x-is-a-varref )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "(define (id x) x)"))
     ==> (define (id x) x) )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "(lambda (x) x)"))
     ==> (lambda (x) x) )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "(lambda x x)"))
     ==> (lambda x x) )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "(lambda x (define y y) x)"))
     ==> (lambda x (define y y) x) )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "(lambda x (define y y) (define (foo z . q) q) 1 2 x)"))
     ==> (lambda x (define y y) (define (foo z . q) q) (begin 1 2 x)) )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "(define list (lambda x (define y y)
                               (define (foo z . q) q)
                               (begin 1 2 x)))"))
     ==> (define list (lambda x (define y y)
			        (define (foo z . q) q)
				(begin 1 2 x))) )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "(define foo (lambda (x) x))"))
     ==> (define foo (lambda (x) x)) )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "(define bar (lambda (e1 e2 e3) (if e1 e2 e3)))"))
     ==> (define bar (lambda (e1 e2 e3) (if e1 e2 e3))) )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "(define bar (lambda (e1 e2 e3) (if e1 e2)))"))
     ==> (define bar (lambda (e1 e2 e3) (if e1 e2))) )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "(define b (begin e1 e2 e3 e4))"))
     ==> (define b (begin e1 e2 e3 e4)) )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "(let ((x 1) (y 2)) x y)"))
     ==> (let ((x 1) (y 2)) (begin x y)) )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "(let ((x 1) (y 2))
           (define a 1)
           (define b (lambda (z) z))
           a x y)"))
     ==> (let ((x 1) (y 2))
	   (define a 1)
	   (define b (lambda (z) z))
	   (begin a x y)) )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "(letrec ((id (lambda (x) x))
                 (id2 (lambda (y) y)))
         (define a 1) (define b (lambda (z) z))
          a x y)"))
     ==> (letrec ((id (lambda (x) x))
		  (id2 (lambda (y) y)))
	   (define a 1) (define b (lambda (z) z))
	   (begin a x y)) )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "(list 3 4)"))
     ==> (list 3 4) )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "(newline)"))
     ==> (newline) )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "()"))
     ==> () )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "(let* ((x 1)
               (y (+ x 2)))
          (define a (lambda (x) 3))
          (define b (lambda (z) z))
          (a x y))"))
     ==> (let* ((x 1)
		(y (+ x 2)))
	   (define a (lambda (x) 3))
	   (define b (lambda (z) z))
	   (a x y)) )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "(set! x 3)"))
     ==> (set! x 3) )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "(and #t (or #f #t #f (and) (or) #t))"))
     ==> (and #t (or #f #t #f (and) (or) #t)) )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "(force (delay 3))"))
     ==> (force (delay 3)) )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "(has-type number 3)"))
     ==> (has-type number 3) )

   (  (tc:unparse-scheme-form
       (tc:character-string-form-parser
	"(has-type (list t) ())"))
      ==> (has-type (list t) ()) )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "(has-type number 3)"))
     ==> (has-type number 3) )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "(has-type-trusted (forall (t) (-> (t s) t)) f)"))
     ==> (has-type-trusted (forall (t) (-> (t s) t)) f) )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "(has-type-trusted (forall (s t) (-> (t s) t)) (lambda (x y) x))"))
     ==> (has-type-trusted (forall (s t) (-> (t s) t)) (lambda (x y) x)) )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "(test-type number? 3)"))
     ==> (test-type number? 3) )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "(test-type (list-of boolean?) (list #t #f))"))
     ==> (test-type (list-of boolean?) (list #t #f)) )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "(deftype newline (-> () void))"))
     ==> (deftype newline (-> () void)) )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "(define-datatype hmm hmm? (nothing))"))
     ==> (define-datatype hmm hmm? (nothing)) )

   ( (tc:unparse-scheme-form
      (tc:character-string-form-parser
       "(define-datatype car car? (auto (name (-> () number?))
                                    (make symbol?) (weight number?)))"))
     ==> (define-datatype car car? (auto (name (-> () number?))
					 (make symbol?) (weight number?))) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "(deftype foo number)
         (define-datatype car car? (auto (name (-> (number?) number?))
                                         (make symbol?) (weight number?)))
         3
         (define foo 1)"))
     ==> (begin
	   (deftype foo number)
	   (define-datatype car car? (auto (name (-> (number?) number?))
					   (make symbol?) (weight number?)))
	   3
	   (define foo 1)) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "(deftype foo number)
         (define-datatype car car? (auto (name (-> (number? number? ...) number?))
                                         (make symbol?) (weight number?)))
         3
         (define foo 1)"))
     ==> (begin
	   (deftype foo number)
	   (define-datatype car car? (auto (name (-> (number? number? ...) number?))
					   (make symbol?) (weight number?)))
	   3
	   (define foo 1)) )


   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "(begin (deftype a number) (define a 1) (define b 2))"))
     ==> (begin (deftype a number) (define a 1) (define b 2)) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "(begin (deftype a number) (define a 1) (set! a 3) (define b 2))"))
     ==> (begin (deftype a number) (define a 1) (set! a 3) (define b 2)) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "(define-datatype vehicle vehicle?
         (auto (make symbol?) (year number?))
         (truck (make symbol?) (year number?) (tonnage number?)))"))
     ==> (define-datatype vehicle vehicle?
	   (auto (make symbol?) (year number?))
	   (truck (make symbol?) (year number?) (tonnage number?))) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "(define (foo x)
           (begin (deftype a number)
                  (define a 1)
                  (define b 2))
           (begin (set! x 1)
                  x))"))
     ==> (define (foo x)
	   (deftype a number)
	   (define a 1)
	   (define b 2)
	   (begin (set! x 1)
		  x)) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "'x '1 '#t '\"a string\" `#\\c"))
     ==> (begin 'x '1 '#t '"a string" `#\c) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "'lambda `if"))
     ==> (begin 'lambda `if) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "'() `(x y 3) '(1 2 . 3)"))
     ==> (begin '() `(x y 3) '(1 2 . 3)) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "'#() `#(x y 3)"))
     ==> (begin '#() `#(x y 3)) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "''x '('a b 'c)"))
     ==> (begin ''x '('a b 'c)) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "`,x `(,y ,@ls `(a ,z q `r))"))
     ==> (begin `,x `(,y ,@ls `(a ,z q `r))) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "`,x `(,y ',m ,@ls `(a ,z ,,a ,@m ,,@(list `(,q)) q `r))
        (quasiquote ((unquote y)
                     (quote (unquote m))
                     (unquote-splicing ls)
                     (quasiquote (a (unquote z) (unquote (unquote a))
                                    (unquote-splicing m)
                                    (unquote (unquote-splicing
                                             (list (quasiquote ((unquote q))))))
                                    q
                                    (quasiquote r)))))"))
     ==> (begin
	   `,x `(,y ',m ,@ls `(a ,z ,,a ,@m ,,@(list `(,q)) q `r))
	   (quasiquote ((unquote y)
			(quote (unquote m))
			(unquote-splicing ls)
			(quasiquote (a (unquote z) (unquote (unquote a))
				       (unquote-splicing m)
				       (unquote (unquote-splicing
						 (list (quasiquote ((unquote q))))))
				       q
				       (quasiquote r)))))) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "(define (length ls)
           (cond
            ((null? ls) 0)
            (else (+ 1 (length (cdr ls))))))"))
     ==> (define (length ls)
	   (cond
	    ((null? ls) 0)
	    (else (+ 1 (length (cdr ls)))))) )


   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "(cond
           ((and (null? d) (= 0 (length d)))
            (display \"hooray\") (newline))
           ((null? d) (display \"was-null\") (newline))
           (#f (error 'shouldnothappen)))
         (cond
           ((null? ls))
           ((foobar x) => f)
           (else (+ 1 (length (cdr ls)))))"))
     ==>(begin
	  (cond
	   ((and (null? d) (= 0 (length d)))
	    (begin (display "hooray") (newline)))
	   ((null? d) (begin (display "was-null") (newline)))
	   (#f (error 'shouldnothappen)))
	  (cond
	   ((null? ls))
	   ((foobar x) => f)
	   (else (+ 1 (length (cdr ls)))))) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "(case (+ 5 6)
          ((1 2 3) (display \"hooray\") (newline) 25)
          ((a #\\x hmm) 3)
          (() 0)
          (else 99))"))
     ==> (case (+ 5 6)
	   ((1 2 3) (begin (display "hooray") (newline) 25))
	   ((a #\x hmm) 3)
	   (() 0)
	   (else 99)) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "(case (+ 5 6)
          ((1 2 3) (display \"hooray\") (newline))
          ((a #\\x hmm) (display 'hmm)))"))
     ==> (case (+ 5 6)
	   ((1 2 3) (begin (display "hooray") (newline)))
	   ((a #\x hmm) (display 'hmm))) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "(cases lambda-1-exp exp
          (lit (datum) datum)
          (varref () (displayln 'varref-case) 'a-varref)
          (lambda (formals body) `(lambda ,formals ,body))
          (else (error 'shouldnthappen)))
        (cases lambda-if-exp exp
          (lit (datum) (set! x 'datum))
          (if (test-exp then-exp else-exp) (set! x then-exp)))"))
     ==> (begin
	   (cases lambda-1-exp exp
	     (lit (datum) datum)
	     (varref () (begin (displayln 'varref-case) 'a-varref))
	     (lambda (formals body) `(lambda ,formals ,body))
	     (else (error 'shouldnthappen)))
	   (cases lambda-if-exp exp
	     (lit (datum) (set! x 'datum))
	     (if (test-exp then-exp else-exp) (set! x then-exp)))) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "(quote lambda) (quasiquote if) (quote (quote x))"))
     ==> (begin (quote lambda) (quasiquote if) (quote (quote x))) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "(do ((vec (make-vector 5))
             (i 0 (+ i 1)))
            ((= i 5) vec)
          (display \"in body\")
          (newline)
          (vector-set! vec i i))
       (let ((x '(1 3 5 7 9)))
          (do ((x x (cdr x))
               (sum 0 (+ sum (car x))))
              ((null? x) sum)))
       (do () (#f))"))
     ==>(begin
	  (do ((vec (make-vector 5))
	       (i 0 (+ i 1)))
	      ((= i 5) vec)
	    (display "in body")
	    (newline)
	    (vector-set! vec i i))
	  (let ((x '(1 3 5 7 9)))
	    (do ((x x (cdr x))
		 (sum 0 (+ sum (car x))))
		((null? x) sum)))
	  (do () (#f))) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "(do ((i (+ -1 (vector-length ht)) (+ -1 i)))
            ((negative? i) nht)
          (vector-set! nht i (alist-map proc (vector-ref ht i))))"))
     ==> (do ((i (+ -1 (vector-length ht)) (+ -1 i)))
	     ((negative? i) nht)
	   (vector-set! nht i (alist-map proc (vector-ref ht i)))) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "(do ((i (+ -1 (vector-length ht)) (+ -1 i)))
            ((negative? i))
          (alist-for-each proc (vector-ref ht i)))"))
     ==> (do ((i (+ -1 (vector-length ht)) (+ -1 i)))
	     ((negative? i))
	   (alist-for-each proc (vector-ref ht i))) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "(define-syntax deftype
          (syntax-rules ()
           ((deftype name type)
            \"#<unspecified>\")))"))
     ==> (define-syntax deftype
	   (syntax-rules ()
	     ((deftype name type)
	      "#<unspecified>"))) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "(define-syntax when
          (syntax-rules (do)
    	    ((when test do stmt1 stmt2 ...)
    	      (if test
    	          (begin stmt1
    		         stmt2 ...)))))"))
     ==> (define-syntax when
	   (syntax-rules (do)
	     ((when test do stmt1 stmt2 ...)
	      (if test
		  (begin stmt1
			 stmt2 ...))))) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "(let-syntax ((when
		     (syntax-rules (do)
		       ((when test do stmt1 stmt2 ...)
			(if test
			    (begin stmt1
				   stmt2 ...))))))
	 (when #t do (displayln 'it-worked) (newline)))"))
     ==> (let-syntax ((when
		       (syntax-rules (do)
			 ((when test do stmt1 stmt2 ...)
			  (if test
			      (begin stmt1
				     stmt2 ...))))))
	   (when #t do (displayln 'it-worked) (newline))) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "(letrec-syntax
           ((my-or (syntax-rules ()
                     ((my-or) #f)
                     ((my-or e) e)
                     ((my-or e1 e2 ...)
                      (let ((temp e1))
                        (if temp
                            temp
                            (my-or e2 ...)))))))
           (let ((x #f)
                 (y 7)
                 (temp 8)
                 (my-let odd?)
                 (my-if even?))
             (my-or x
                    (my-let temp)
                    (my-if y)
                    y)))"))
     ==> (letrec-syntax
	     ((my-or (syntax-rules ()
		       ((my-or) #f)
		       ((my-or e) e)
		       ((my-or e1 e2 ...)
			(let ((temp e1))
			  (if temp
			      temp
			      (my-or e2 ...)))))))
	   (let ((x #f)
		 (y 7)
		 (temp 8)
		 (my-let odd?)
		 (my-if even?))
	     (my-or x
		    (my-let temp)
		    (my-if y)
		    y))) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "(has-type (forall (s t) (-> ((-> (s) t) (list-of s)) (list-of t)))
                  map)"))
     ==> (has-type (forall (s t) (-> ((-> (s) t) (list-of s)) (list-of t)))
		   map) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "(has-type (forall (s t) (-> ((-> (s ...) t) (list-of s) ...) (list-of t)))
                  map)"))
     ==> (has-type (forall (s t) (-> ((-> (s ...) t) (list-of s) ...) (list-of t)))
		   map) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "(has-type (all-of (list-of number) (list-of boolean)) '())"))
     ==> (has-type (all-of (list-of number) (list-of boolean)) '()) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "(-> (char? number?) number?)
        (forall (t) t)"))
     ==> (begin (-> (char? number?) number?)
		(forall (t) t)) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "(let loop ((numbers '(3 -2 1 6 -5))
                   (nonneg '())
                   (neg '()))
             (cond
                ((null? numbers) (list nonneg neg))
                ((>= (car numbers) 0)
                 (loop (cdr numbers)
                       (cons (car numbers) nonneg)
                       neg))
                ((< (car numbers) 0)
                 (loop (cdr numbers)
                       nonneg
                       (cons (car numbers) neg)))))"))
     ==> (let loop ((numbers '(3 -2 1 6 -5))
		    (nonneg '())
		    (neg '()))
	   (cond
	    ((null? numbers) (list nonneg neg))
	    ((>= (car numbers) 0)
	     (loop (cdr numbers)
		   (cons (car numbers) nonneg)
		   neg))
	    ((< (car numbers) 0)
	     (loop (cdr numbers)
		   nonneg
		   (cons (car numbers) neg))))) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "(define-datatype pi-calc-expr pi-calc-expr?
          (parallel-composition (list-of pi-calc-expr?))
          (restriction (name symbol?) (body pi-calc-expr?))
          (input (name symbol?) (formals (list-of symbol?)) (body pi-calc-expr?))
          (output (name symbol?) (actuals (list-of symbol?)))
          (replication (body pi-calc-expr?))
          (null ))
         (has-type
           (variant-record
             (parallel-composition (exprs (list-of pi-calc-expr)))
             (restriction (name symbol) (body pi-calc-expr))
             (input (name symbol) (formals (list-of symbol)) (body pi-calc-expr))
             (output (name symbol) (actuals (list-of symbol)))
             (replication (body pi-calc-expr))
             (null ))
            pi-calc-expr)"
       ))
     ==> (begin
	   (define-datatype pi-calc-expr pi-calc-expr?
	     (parallel-composition (list-of pi-calc-expr?))
	     (restriction (name symbol?) (body pi-calc-expr?))
	     (input (name symbol?) (formals (list-of symbol?)) (body pi-calc-expr?))
	     (output (name symbol?) (actuals (list-of symbol?)))
	     (replication (body pi-calc-expr?))
	     (null ))
	   (has-type
	     (variant-record
	      (parallel-composition (exprs (list-of pi-calc-expr)))
	      (restriction (name symbol) (body pi-calc-expr))
	      (input (name symbol) (formals (list-of symbol)) (body pi-calc-expr))
	      (output (name symbol) (actuals (list-of symbol)))
	      (replication (body pi-calc-expr))
	      (null ))
	     pi-calc-expr)) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "(has-type (type-predicate-for number) number?)"))
     ==> (has-type (type-predicate-for number) number?) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "(module module-a (lib \"typedscm.ss\" \"typedscm\")
           (provide x) (define x 3))"))
     ==> (module module-a (lib "typedscm.ss" "typedscm")
	   (provide x) (define x 3)) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "(require (lib \"module-a.scm\" \"typedscm\"))"))
     ==> ;; [[[FIXME: is the following what we want?]]]
     (begin (require (lib "module-a.scm" "typedscm"))) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "(require (lib \"module-a.scm\" \"typedscm\"))
        (require (lib \"module-b.scm\" \"typedscm\"))"))
     ==> (begin (require (lib "module-a.scm" "typedscm"))
		(require (lib "module-b.scm" "typedscm"))) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "(module module-b (lib \"typedscm.ss\" \"typedscm\")
          (provide x y)
          (require (lib \"module-a.scm\" \"typedscm\"))
          (define y 3)
          )"))
     ==> (module module-b (lib "typedscm.ss" "typedscm")
	   (provide x y)
	   (require (lib "module-a.scm" "typedscm"))
	   (define y 3)
	   ) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "(module module-c (lib \"typedscm.ss\" \"typedscm\")
          (provide x y)
          (defrep (foo bar))
          (define y 3)
          )"))
     ==> (module module-c (lib "typedscm.ss" "typedscm")
	   (provide x y)
	   (defrep (foo bar))
	   (define y 3)
	   ) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "(module module-c2 (lib \"typedscm.ss\" \"typedscm\")
          (provide x y)
          (defrep (forall (t u) (my-pair-of t u) (pair-of t u))
                  (forall (t) (stack-of t) (list-of t)))
          )"))
     ==> (module module-c2 (lib "typedscm.ss" "typedscm")
	   (provide x y)
           (defrep (forall (t u) (my-pair-of t u) (pair-of t u))
	           (forall (t) (stack-of t) (list-of t)))
	   ) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "(module module-d (lib \"typedscm.ss\" \"typedscm\")
          (provide x y z w)
          (require module-a \"module-b.scm\"
                   (file \"test/module-c.scm\")
                   (lib \"module-c.scm\" \"typedscm\" \"test\"))
          (defrep (foo bar) (baz (list-of string)))
          (define y 3)
          (define z (lambda (x) x))
          (define w (lambda (f) (lambda (g) (lambda (x) (f (g x))))))
          )"))
     ==>
     (module module-d (lib "typedscm.ss" "typedscm")
       (provide x y z w)
       (require module-a "module-b.scm"
		(file "test/module-c.scm")
		(lib "module-c.scm" "typedscm" "test"))
       (defrep (foo bar) (baz (list-of string)))
       (define y 3)
       (define z (lambda (x) x))
       (define w (lambda (f) (lambda (g) (lambda (x) (f (g x))))))
       ) )

   ( (tc:unparse-scheme-program
      (tc:character-string-program-parser
       "(module module-d (lib \"typedscm.ss\" \"typedscm\")
          (provide x y z w)
          (deftype x number)
          (deftype y number)
          (deftype z (-> (number) number))
          (require module-a \"module-b.scm\"
                   (file \"test/module-c.scm\")
                   (lib \"module-c.scm\" \"typedscm\" \"test\"))
          (defrep (foo bar) (baz (list-of string)))
          (define y 3)
          (define z (lambda (x) x))
          (deftype w
             (forall (s t u) (-> ((-> (s) t)) (-> ((-> (u) s)) (-> (u) t)))))
          (define w (lambda (f) (lambda (g) (lambda (x) (f (g x))))))
        )"))
     ==>
     (module module-d (lib "typedscm.ss" "typedscm")
       (provide x y z w)
       (deftype x number)
       (deftype y number)
       (deftype z (-> (number) number))
       (deftype w
	 (forall (s t u) (-> ((-> (s) t)) (-> ((-> (u) s)) (-> (u) t)))))
       (require module-a "module-b.scm"
		(file "test/module-c.scm")
		(lib "module-c.scm" "typedscm" "test"))
       (defrep (foo bar) (baz (list-of string)))
       (define y 3)
       (define z (lambda (x) x))
       (define w (lambda (f) (lambda (g) (lambda (x) (f (g x))))))
       ) )
   )
 )
