;;; $Id: tc-tree.scm,v 1.12 2006/01/18 23:17:27 leavens Exp $
;;; Copyright (C) 2006 Iowa State University
;;;
;;; This file is part of Typedscm.
;;;
;;; This library is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU Lesser General Public License
;;; as published by the Free Software Foundation; either version 2.1,
;;; of the License, or (at your option) any later version.
;;;
;;; This library is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; Lesser General Public License for more details.
;;;
;;; You should have received a copy of the GNU Lesser General Public
;;; License along with Typedscm; see the file LesserGPL.txt.  If not,
;;; write to the Free Software Foundation, Inc., 51 Franklin St, Fifth
;;; Floor, Boston, MA 02110-1301 USA.

;;; One way to do trees in Scheme
;;;
;;; AUTHOR: Gary T. Leavens, Brian Dorn

(module tc-tree (lib "typedscm.ss" "typedscm")

(provide tc:tree-of tc:root tc:leaves tc:node tc:print-map tc:tree-print)

(deftype tc:tree-of (forall (a)
		      (-> ((type-predicate-for a))
			  (type-predicate-for (tc:tree-of a)))))
(deftype tc:root (forall (a) (-> ((tc:tree-of a)) a)))
(deftype tc:leaves (forall (a) (-> ((tc:tree-of a))
				(list-of (tc:tree-of a)))))
(deftype tc:node (forall (a) (-> (a (list-of (tc:tree-of a))) (tc:tree-of a))))

(defrep (forall (a) (tc:tree-of a) (pair-of a (list-of (tc:tree-of a)))))

(define tc:tree-of
  (has-type-trusted
    (forall (a)
      (-> ((type-predicate-for a))
	  (type-predicate-for (tc:tree-of a))))
  (forall (elemtype?)
    (lambda (val)
      ((pair-of elemtype?
		(list-of (tc:tree-of elemtype?)))
       val)))))

(define tc:root
  (lambda (tree)
    (car tree)))

(define tc:leaves
  (has-type-trusted
    (forall (a) (-> ((pair-of a (list-of (tc:tree-of a))))
		    (list-of (pair-of a (list-of (tc:tree-of a))))))
  (lambda (tree)
    (cdr tree))))

(define tc:node
  (has-type-trusted
    (forall (a) (-> (a (list-of (tc:tree-of a)))
		    (pair-of a (list-of (tc:tree-of a)))))
  (lambda (a as)
    (cons a as))))

(define tc:print-map
  (lambda (printer l)
    (if (not (null? l))
	(begin
	  (printer (car l))
	  (tc:print-map printer (cdr l))))))

(define tc:tree-print
  (lambda (tree)
    (tree-print-help tree "")))

(define tree-print-help
  (lambda (tree prefix)
    (displayln prefix "-" (tc:root tree))
    (leaf-print (tc:leaves tree) (string-append prefix " |"))))

(define leaf-print
  (lambda (ls prefix )
    (tc:print-map (lambda (t) (tree-print-help t prefix)) ls)))

) ;; end module
