;;; $Id: tc-types.scm,v 1.97 2006/04/04 17:54:10 leavens Exp $
;;; Copyright (C) 2006 Iowa State University
;;;
;;; This file is part of Typedscm.
;;;
;;; This library is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU Lesser General Public License
;;; as published by the Free Software Foundation; either version 2.1,
;;; of the License, or (at your option) any later version.
;;;
;;; This library is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; Lesser General Public License for more details.
;;;
;;; You should have received a copy of the GNU Lesser General Public
;;; License along with Typedscm; see the file LesserGPL.txt.  If not,
;;; write to the Free Software Foundation, Inc., 51 Franklin St, Fifth
;;; Floor, Boston, MA 02110-1301 USA.

;;; Author: Brian Dorn and Gary T. Leavens

(module tc-types (lib "typedscm.ss" "typedscm")

(provide
 tc:type-expr tc:type-expr?
 tc:basic-type-expr tc:function-type-expr tc:intersection-type-expr
 tc:type-predicate-for-type-expr tc:variant-record-type-expr
 tc:applied-type-expr tc:variant-type-expr tc:field-type-binding-type-expr
 tc:variable-type-expr tc:declaration-type-expr tc:error-type-expr
 ;; Procedures for Unifiable Method Dictionary
 tc:app tc:app-list tc:type-expr-same-kind
 tc:get-name tc:list-same-kind tc:simple-same-kind tc:get-kind
 tc:type-expr-subterms tc:type-expr-contravar-subterms
 tc:type-expr-covar-subterms tc:type-expr-invar-subterms
 tc:type-expr-subtype-replace tc:intersection-type-expr?
 tc:find-intersection-subtyping tc:find-intersection-supertyping
 tc:poof-type-expr? tc:datum-type-expr? tc:error-type-expr?
 tc:declaration-type-expr? tc:declaration-type-expr->env
 tc:type-expr-null-subst tc:type-expr-bind tc:type-expr-as-unifiable
 tc:make-variable-type-expr tc:new-variable-type-expr
 tc:listof-new-variable-type-expr tc:make-basic-type-expr
 tc:make-function-type-expr tc:make-intersection-type-expr
 tc:make-type-predicate-for-type-expr tc:make-variant-record-type-expr
 tc:make-variant-type-expr tc:make-variant-type-decl-type-expr
 tc:make-field-type-binding-type-expr tc:make-applied-type-expr
 ;; Type Testers
 tc:variable-type-expr? tc:basic-type-expr?
 tc:non-variable-basic-type-expr? tc:dots-type-expr?
 tc:function-type-expr? tc:varargs-function-type-expr? tc:monotype?
 tc:type-predicate-for-type-expr? tc:variant-record-type-expr?
 tc:variant-type-expr? tc:field-type-binding-type-expr?
 tc:applied-type-expr? tc:contains-function-type-expr?
 ;; Extraction Helpers
 tc:variable-type-expr->uniq-sym tc:uniq-sym->variable-type-expr
 tc:basic-type-expr->symbol
 tc:function-type-expr->arg-types tc:function-type-expr->result-type
 tc:intersection-type-expr->conjoined-types
 tc:type-predicate-for-type-expr->type
 tc:variant-type-expr->variant-name tc:variant-type-expr->fields
 tc:applied-type-expr->operator-type tc:applied-type-expr->operand-types
 tc:variant-record-type-expr->variants
 tc:field-type-binding-type-expr->field-name
 tc:field-type-binding-type-expr->type
 tc:error-type-expr->error-record
 tc:type-expr-replace
 ;; OTHER HELPERS
 tc:basic-type-expr-eq? tc:uniq-sym->symbol
 tc:type-expr-instance tc:variable-type-expr-new-level tc:type-expr-uniquify
 tc:fix-length
 ;; ERROR RECORDS
 tc:error-record tc:error-record?
 tc:module-missing-deftype-error-record 
 tc:mismatch-error-record tc:composite-error-record
 tc:if-subtype-error-record tc:apply-non-proc-error-record
 tc:op-arg-mismatch-error-record tc:varref-error-record
 tc:badtest-error-record tc:defined-lambda-mismatch-error-record
 tc:named-lambda-subtyping-error-record
 tc:duplicate-let-binding-error-record tc:cases-clause-bad-fields-error-record
 tc:cases-on-non-variant-error-record
 tc:cases-with-bad-clause-usage-error-record
 tc:duplicate-formals-error-record tc:non-void-body-error-record
 tc:set!-mismatch-error-record tc:cond-clauses-mismatch-error-record
 tc:cases-clauses-mismatch-error-record tc:has-type-mismatch-error-record
 tc:forall-bad-body-error-record tc:define-type-mismatch-error-record
 tc:define-datatype-name-conflict-error-record
 tc:unknown-module-error-record tc:bad-file-error-record
 tc:deftype-redeclaration-error-record tc:simple-error-record
 tc:mismatch-error-maker tc:rule-error-maker
 tc:scheme-definition-error-maker tc:scheme-formals-error-maker
 tc:scheme-exp-error-maker tc:scheme-cond-clause-error-maker
 tc:generic-error-maker tc:composite-error-maker
 tc:varref-error-maker tc:unknown-module-error-maker
 tc:defined-lambda-mismatch-error-maker
 tc:named-lambda-subtyping-error-maker
 tc:duplicate-let-binding-error-maker
 tc:cases-clause-bad-fields-error-maker
 tc:cases-on-non-variant-record-error-maker
 tc:cases-with-bad-clause-usage-error-maker
 tc:bad-file-error-maker
 tc:simple-error-maker tc:mismatch-error-record?
 tc:mismatch-error-record->inferred-type
 tc:mismatch-error-record->expected-type
 tc:composite-error-record? tc:contains-errors? tc:extract-errors
 ;; TYPE BINDINGS
 tc:binding? tc:bind-type-expr tc:bound?
 tc:binding->type-expr tc:binding->binder
 ;; all-from tc-subst.scm
 tc:make-subst tc:subst? tc:subst-compose
 tc:subst-apply tc:make-null-subst
 tc:unifiable-md tc:unifiable-md? tc:unifiable-md-dict
 tc:make-unifiable-md
 tc:mgu tc:subtype-mgu tc:vars-in
 tc:list-unify tc:list-subtype-unify tc:rename-vars
 ;; all-from tc-logical-variables.scm
 tc:logicalvar? tc:variable? tc:make-logicalvar
 tc:symbol->logicalvar tc:new-logicalvar tc:new-lv-list
 tc:logicalvar->name tc:logicalvar->level tc:variable-eq?
 tc:logicalvar->uniq-sym
 ;; all-from tc-attrib-pair.scm (except :)
 tc:attrib-pair-of tc:attrib-pair->syn tc:attrib-pair->type
 tc:tree-of tc:root tc:leaves tc:node 
 tc:print-map tc:tree-print
 ;; all-from tc-env.scm
 tc:env-of tc:env-empty tc:env-update tc:env-update* tc:env-join tc:env-join*
 tc:env-bound? tc:env-value tc:env-map tc:env-filter tc:env-for-each
 tc:assoc-list->env
 tc:environment? tc:env-names tc:env-vals
; chain-count resolve-count total-app-count no-chain
 )

(require (lib "tc-util.scm" "typedscm")
	 (lib "tc-subst.scm" "typedscm")
         (lib "maybe.scm" "typedscm")
	 (lib "relax-type-checking.scm" "typedscm")
	 (lib "tc-tracing.scm" "typedscm")
	 (lib "tc-env.scm" "typedscm")
         (lib "tc-position.scm" "typedscm")
         (lib "tc-scheme-abstract-syntax.scm" "typedscm"))

(defrep (tc:binding (pair-of symbol tc:type-expr))
        (uniq-sym symbol))

;; We model type expressions using the following datatype definition.

(define-datatype tc:type-expr tc:type-expr?
  (tc:basic-type-expr (symbol symbol?))
  (tc:function-type-expr
   (arg-types (list-of tc:type-expr?)) (result-type tc:type-expr?))
  (tc:intersection-type-expr (conjoined-types (list-of tc:type-expr?)))
  (tc:type-predicate-for-type-expr (type tc:type-expr?))
  (tc:variant-record-type-expr
   (variants (list-of tc:type-expr?))
   ;; where variants are tc:variant-type-decl-type-exprs
   )
  (tc:applied-type-expr
   (operator-type tc:type-expr?) (operand-types (list-of tc:type-expr?)))
  ;; The following are to facilitate type checking
  ;; and aren't part of the type expression grammar seen by users
  (tc:variant-type-expr
   (variant-name symbol?)
   (fields (list-of tc:type-expr?))
   ;; where fields are tc:field-type-binding-type-exprs
   )
  (tc:field-type-binding-type-expr
   (field-name symbol?) (type tc:type-expr?))
  (tc:variable-type-expr (lvar tc:logicalvar?))
  ;; modules
  (tc:declaration-type-expr (pi (tc:env-of symbol? tc:binding?)))
  ;; errors
  ;; [[[FIXME: add repaired-error-type-expr for type error recovery]]]
  (tc:error-type-expr (error-record tc:error-record?))
  )


;;          Procedures for Unifiable Method Dictionary
;;
;;  These procedures enable tc:type-expr to be unified using
;;  tc-subst.

;(define chain-count 0)
;(define resolve-count 0)
;(define total-app-count 0)
;(define no-chain 0)

(deftype tc:app (-> (tc:subst)
		    (-> (tc:type-expr) tc:type-expr)))
(define tc:app
  (lambda (s)
    (lambda (attrib)

 ;     (set! total-app-count (+ 1 total-app-count))
      
      ;; ENSURES: result is a function that maps type expressions
      ;;          to type expressions based on the substitution
      (cases tc:type-expr attrib
	     (tc:function-type-expr (arg-types result-type)
				    (tc:function-type-expr
				     ((tc:app-list s) arg-types)
				     ((tc:app s) result-type)))
	     (tc:intersection-type-expr (conjoined-types)
					(tc:intersection-type-expr
					 ((tc:app-list s) conjoined-types)))
	     (tc:type-predicate-for-type-expr (type)
					      (tc:type-predicate-for-type-expr
					       ((tc:app s) type)))
	     (tc:variant-record-type-expr (variants)
					  (tc:variant-record-type-expr
					   ((tc:app-list s) variants)))
	     (tc:applied-type-expr (operator-type operand-types)
				   (tc:applied-type-expr
				    ((tc:app s) operator-type)
				    ((tc:app-list s) operand-types)))
	     (tc:variable-type-expr (lvar)
;				    (set! chain-count 0)
				    (let ((result (tc:resolve-lvar s lvar '())))
;				      (if (> chain-count 0)
;					  (displayln "Chain len: " chain-count)
;					  (set! no-chain (+ 1 no-chain)))			      
				      (if (tc:variable-type-expr? result)
					  result
					  ((tc:app s) result))))
	     (tc:declaration-type-expr (pi)
				       (tc:declaration-type-expr
					(tc:env-map
					 (lambda (key binding)
					   (tc:bind-type-expr
					    (tc:binding->binder binding)
					    ((tc:app s) (tc:binding->type-expr binding))))	    
					 pi)))
	     (tc:variant-type-expr (variant-name fields)
				   (tc:variant-type-expr
				    variant-name
				    ((tc:app-list s) fields)))
	     (tc:field-type-binding-type-expr (field-name type)
					      (tc:field-type-binding-type-expr
					       field-name
					       ((tc:app s) type)))
	     (else attrib)))))

(deftype tc:resolve-lvar
  (-> (tc:subst tc:logicalvar (list-of tc:logicalvar)) tc:type-expr))
(define tc:resolve-lvar
  (lambda (s lvar lvar-list)
;    (set! resolve-count (+ 1 resolve-count))
    (let ((result (tc:subst-apply s lvar)))
      (if (and (tc:variable-type-expr? result)
	       (not (equal? result (tc:variable-type-expr lvar))))	  
	  (cases tc:type-expr result
		 (tc:variable-type-expr
		  (internal-lvar)
;		  (set! chain-count (+ chain-count 1))
		  (tc:resolve-lvar s internal-lvar (cons lvar lvar-list)))
		 (else (error "Bad type expression in subst-apply-fp")))
	  result))))

(deftype tc:app-list
  (-> (tc:subst)
      (-> ((list-of tc:type-expr)) (list-of tc:type-expr))))
(define tc:app-list
  (lambda (s)
    (let ((app (tc:app s)))
      (lambda (ls)
	(map app ls)))))

(deftype tc:type-expr-same-kind (-> (tc:type-expr tc:type-expr) boolean))
(define tc:type-expr-same-kind
  (lambda (t1 t2)
    ;; ENSURES: result is true when t1 and t2 are the same 'kind' of term
    (cases tc:type-expr t1	   
	   (tc:basic-type-expr
	    (sym1)
	    ;; added this comparison to correct behavior
	    (and (eq? (tc:get-kind t2)
		      'basic)
		 (eq? sym1 (tc:basic-type-expr->symbol t2)))) 
	   (tc:declaration-type-expr
	    (pi1) 
	    (cases tc:type-expr t2
	      (tc:declaration-type-expr
	       (pi2) 
	       (equal?
		(tc:env-names pi1)
		(tc:env-names pi2)))
	      (else (tc:simple-same-kind t1 t2))))
	   (tc:variant-record-type-expr
	    (variants)
	    (and (eq? (tc:get-kind t2)
		      'variant-record)
		 (tc:list-same-kind
		  variants
		  (tc:variant-record-type-expr->variants t2))))
	   (tc:variant-type-expr
	    (variant-name fields)
	    (and (eq? (tc:get-kind t2)
		      'variant)
		 (eq? variant-name
		      (tc:get-name t2))))	   
	   (tc:field-type-binding-type-expr (field-name type)
					    (and (eq? (tc:get-kind t2)
						      'field-type-binding)
						 (eq? field-name
						      (tc:get-name t2))))
	   (else (tc:simple-same-kind t1 t2)))))

(deftype tc:get-name (-> (tc:type-expr) symbol))
(define tc:get-name
  (lambda (te)
    ;; ENSURES: result is the name value for a type-expr when such
    ;;          a field is present
    (cases tc:type-expr te
	   (tc:variant-type-expr (variant-name fields) variant-name)
	   (tc:field-type-binding-type-expr (field-name type) field-name)
	   (else (error "Shouldn't call get-name on non-record/binding: " te)))))

(deftype tc:list-same-kind
  (-> ((list-of tc:type-expr) (list-of tc:type-expr)) boolean))
(define tc:list-same-kind
  (lambda (vl1 vl2)
    (cond
     ((null? vl1) (null? vl2))
     ((null? vl2) #f)
     (else (and (tc:type-expr-same-kind (car vl1) (car vl2))
		(tc:list-same-kind (cdr vl1) (cdr vl2)))))))

(deftype tc:simple-same-kind (-> (tc:type-expr tc:type-expr) boolean))
(define tc:simple-same-kind
  (lambda (t1 t2)
    ;; ENSURES: result is true when two type-exprs t1 and t2 have
    ;;          the same kind, without considering any substrutures
    ;;          they might have
    (let ((k1 (tc:get-kind t1)) (k2 (tc:get-kind t2)))
      (eq? k1 k2))))

(deftype tc:get-kind (-> (tc:type-expr) symbol))
(define tc:get-kind
  (lambda (te)
    (cases tc:type-expr te
	   (tc:basic-type-expr (symbol) 'basic)
	   (tc:function-type-expr (arg-types result-type) 'function)
	   (tc:intersection-type-expr (conjoined-types) 'intersection)
	   (tc:type-predicate-for-type-expr (type) 'predicate)
	   (tc:variant-record-type-expr (variants) 'variant-record)
	   (tc:applied-type-expr (operator-type operand-types) 'applied)
	   (tc:variable-type-expr (lvar) 'variable)
	   (tc:declaration-type-expr (pi) 'declaration)
	   (tc:variant-type-expr (variant-name fields) 'variant)
	   (tc:field-type-binding-type-expr (field-name type)
					    'field-type-binding)
	   (tc:error-type-expr (error-record) 'error))))

(deftype tc:type-expr-subterms (-> (tc:type-expr) (list-of tc:type-expr)))
(define tc:type-expr-subterms
  (lambda (te)
    ;; ENSURES: returns a list of the subterms of type-expr te
    (cases tc:type-expr te
	   (tc:function-type-expr (arg-types result-type)
				  (append arg-types (list result-type)))
	   (tc:intersection-type-expr (conjoined-types) conjoined-types)
	   (tc:type-predicate-for-type-expr (type) (list type))
	   (tc:variant-record-type-expr (variants) variants)
	   (tc:applied-type-expr (operator-type operand-types)
				 (cons operator-type operand-types))
	   (tc:declaration-type-expr (pi)
				     (map tc:binding->type-expr (tc:env-vals pi)))
	   (tc:variant-type-expr (variant-name fields) fields)
	   (tc:field-type-binding-type-expr
	    (field-name type) (list type))   
	   (else '())  ;;; basic, error, variable
	   )))

(deftype tc:type-expr-contravar-subterms
  (-> (tc:type-expr) (list-of tc:type-expr)))
(define tc:type-expr-contravar-subterms
  (lambda (te)
    ;; ENSURES: returns a list of subterms which must be treated
    ;;          contravariant in type-expr te
    (cases tc:type-expr te
	   (tc:function-type-expr
	    (arg-types result-type)
	    arg-types)
	   (else '()))))

(deftype tc:type-expr-covar-subterms
  (-> (tc:type-expr) (list-of tc:type-expr)))
(define tc:type-expr-covar-subterms
  (lambda (te)
    ;; ENSURES: returns a list of subterms which must be treated as
    ;;          covariant in type-expr te
    (cases tc:type-expr te
	   (tc:function-type-expr
	    (arg-types result-type)
	    (list result-type))
	   (tc:applied-type-expr
	    (operator-type operand-types)
	    '())
	   (else (tc:type-expr-subterms te)))))

(deftype tc:type-expr-invar-subterms
  (-> (tc:type-expr) (list-of tc:type-expr)))
(define tc:type-expr-invar-subterms
  (lambda (te)
    ;; Ensures: returns a list of subterms which must be treated as
    ;;          invariant in type-expr te
    (cases tc:type-expr te
	   (tc:applied-type-expr
	    (operator-type operand-types)
	    (cons operator-type operand-types))
	   (tc:type-predicate-for-type-expr
	    (type)
	    (list type))
	   (else '()))))

(deftype tc:type-expr-subtype-replace
  (-> (tc:type-expr tc:type-expr) (pair-of tc:type-expr tc:type-expr)))
(define tc:type-expr-subtype-replace
  (lambda (subtype-val supertype-val)
    ;; Replace type predicates with function types
    (if (and (tc:type-predicate-for-type-expr? subtype-val)
	     (tc:function-type-expr? supertype-val))
	(set! subtype-val (tc:function-type-expr
			   (list (tc:basic-type-expr 'datum))
			   (tc:basic-type-expr 'boolean))))
    ;; Replace varargs with functions of the correct length
    (if (and (tc:varargs-function-type-expr? subtype-val)
	     (tc:function-type-expr? supertype-val))
	(set! subtype-val (tc:fix-length subtype-val supertype-val)))
    (if (and (tc:function-type-expr? subtype-val)
	     (tc:varargs-function-type-expr? supertype-val))
	(set! supertype-val (tc:fix-length supertype-val subtype-val)))		
    ;; Build return value
    (cons subtype-val supertype-val)))

(deftype tc:intersection-type-expr? (-> (tc:type-expr) boolean))
(define tc:intersection-type-expr?
  (lambda (d)
    (cases tc:type-expr d
      (tc:intersection-type-expr (conjoined-types) #t)
      (else #f))))

(deftype tc:find-intersection-subtyping
  (-> (tc:type-expr tc:type-expr) (list-of tc:subst)))
(define tc:find-intersection-subtyping
  (lambda (conj-type sought)
    ;; EFFECT: attempts to find a substitution that makes one of the
    ;;         types that "conj-type" has a subtype of "sought".  Tries each
    ;;         conjoined type in conj-type in sequential order, returning the
    ;;         unifiying substitution if one exists.
    (letrec
	((find-solution
	  (lambda (type-list)
	    (if (null? type-list)
		'()  ;; Unification Error
		(let ((subtype-substl ((tc:subtype-mgu (tc:type-expr-as-unifiable))
				       (car type-list) sought)))
		  (if (null? subtype-substl)
		      ;; Move on
		      (find-solution (cdr type-list))
		      subtype-substl))))))

      (find-solution (tc:intersection-type-expr->conjoined-types conj-type)))))

(deftype tc:find-intersection-supertyping
  (-> (tc:type-expr tc:type-expr) (list-of tc:subst)))
(define tc:find-intersection-supertyping
  (lambda (sought conj-type)
    ;; EFFECT: attempts to find a substitution that makes "sought"
    ;;         a subtype of *each* type in "conj-type".
    ;;         Returns the empty list if no substitution exists.
    (letrec ((find-solution
	      (lambda (type-list substl)
		(if (null? type-list)
		    substl
		    (let ((subtype-substl ((tc:subtype-mgu (tc:type-expr-as-unifiable))
					   sought (car type-list))))
		      (if (null? subtype-substl)
			 subtype-substl  ;; Unification Error
			 (find-solution (cdr type-list) 
					(list
					 (tc:subst-compose
					  (car substl) (car subtype-substl))))))))))
      (find-solution (tc:intersection-type-expr->conjoined-types conj-type)
		     (list (tc:type-expr-null-subst)) ))))


(deftype tc:poof-type-expr? (-> (tc:type-expr) boolean))
(define tc:poof-type-expr?
  (lambda (d)
    (cases tc:type-expr d
      (tc:basic-type-expr (symbol) (eq? symbol 'poof))
      (else #f))))

(deftype tc:datum-type-expr? (-> (tc:type-expr) boolean))
(define tc:datum-type-expr?
  (lambda (d)
    (cases tc:type-expr d
      (tc:basic-type-expr (symbol) (eq? symbol 'datum))
      (else #f))))

(deftype tc:error-type-expr? (-> (tc:type-expr) boolean))
(define tc:error-type-expr?
  (lambda (d)
    (cases tc:type-expr d
      (tc:error-type-expr (error-record) #t)
      (else #f))))

(deftype tc:declaration-type-expr? (-> (tc:type-expr) boolean))
(define tc:declaration-type-expr?
  (lambda (d)
    (cases tc:type-expr d
	   (tc:declaration-type-expr (pi) #t)
	   (else #f))))

(deftype tc:declaration-type-expr->env
  (-> (tc:type-expr) (tc:env-of symbol tc:binding)))
(define tc:declaration-type-expr->env
  (lambda (d)
    (cases tc:type-expr d
	   (tc:declaration-type-expr (pi) pi)
	   (else
	    (error "tc:declaration-type-expr->env: not a declaration-type-expr: "
		   d)))))

(deftype tc:type-expr-null-subst (-> () tc:subst))
(define tc:type-expr-null-subst
  (lambda ()
    (tc:make-null-subst tc:variable-type-expr)))

(deftype tc:type-expr-bind (-> (tc:logicalvar tc:type-expr) tc:subst))
(define tc:type-expr-bind
  (lambda (lv typ)
    (tc:make-subst lv typ tc:variable-type-expr)))
    
;; We need to create a method dictionary contianing the appropriate
;; procedures for type expressions so they can be used with
;; substitutions and as unifiable terms.

(deftype tc:type-expr-as-unifiable (-> () tc:unifiable-md))
(define tc:type-expr-as-unifiable
  (lambda ()
    (tc:make-unifiable-md
     ;; to-term
     tc:variable-type-expr
     ;; get-var
     (lambda (te)
       (cases tc:type-expr te
	 (tc:variable-type-expr (v) (make-something v))
	 (else (make-nothing))))
     ;; subterms
     tc:type-expr-subterms
     ;; same-kind
     tc:type-expr-same-kind
     ;; nullSubst
     (tc:type-expr-null-subst)
     ;; bind
     tc:type-expr-bind
     ;; app 
     tc:app
     ;; contravar-subterms
     tc:type-expr-contravar-subterms
     ;; covar-subterms
     tc:type-expr-covar-subterms
     ;; invar-subterms
     tc:type-expr-invar-subterms
     ;; subtype-replace
     tc:type-expr-subtype-replace
     ;; intersection-type?
     tc:intersection-type-expr?
     ;; find-intersection-subtyping
     tc:find-intersection-subtyping
     ;; find-intersection-supertyping
     tc:find-intersection-supertyping
     ;; is-bottom?
     tc:poof-type-expr?
     ;; is-top?
     tc:datum-type-expr?
     ;; is-error?
     tc:error-type-expr?
     ;; gen-error-mismatch
     tc:mismatch-error-maker
     ;; gen-error-rule
     tc:rule-error-maker
     ;; is-dec-type-expr?
     tc:declaration-type-expr?
     ;; dec-type-expr->env
     tc:declaration-type-expr->env
     )))

;;            Helpers to Create Type Expressions

(deftype tc:make-variable-type-expr (-> (symbol) tc:type-expr))
(define tc:make-variable-type-expr
  (lambda (s) (tc:variable-type-expr (tc:symbol->logicalvar s))))

(deftype tc:new-variable-type-expr (-> () tc:type-expr))
(define tc:new-variable-type-expr
  (lambda ()
    ;; EFFECT: returns a unique variable-type-expr
    ;; Note: this has to go through tc:new-variable
    ;; so that no duplication will occur!
    (tc:variable-type-expr (tc:new-logicalvar))))

(deftype tc:listof-new-variable-type-expr (-> (number) (list-of tc:type-expr)))
(define tc:listof-new-variable-type-expr
  (lambda (len)
    (map tc:variable-type-expr
	 (tc:new-lv-list len))))

(deftype tc:make-basic-type-expr (-> (symbol) tc:type-expr))
(define tc:make-basic-type-expr
  (lambda (sym)
    (tc:basic-type-expr sym)))

(deftype tc:make-function-type-expr
  (-> ((list-of tc:type-expr) tc:type-expr) tc:type-expr))
(define tc:make-function-type-expr
  (lambda (lot te)
    (tc:function-type-expr lot te)))

(deftype tc:make-intersection-type-expr
  (-> ((list-of tc:type-expr)) tc:type-expr))
(define tc:make-intersection-type-expr
  (lambda (lot)
    (tc:intersection-type-expr lot)))

(deftype tc:make-type-predicate-for-type-expr
  (-> (tc:type-expr) tc:type-expr))
(define tc:make-type-predicate-for-type-expr
  (lambda (te)
    (tc:type-predicate-for-type-expr te)))

(deftype tc:make-variant-record-type-expr
  (-> ((list-of tc:type-expr)) tc:type-expr))
(define tc:make-variant-record-type-expr
  ;; REQUIRES:  Type expressions in the list must be
  ;;            variant-type-exprs
  (lambda (lot)
    (tc:variant-record-type-expr lot)))

(deftype tc:make-variant-type-expr
  (-> (symbol (list-of tc:type-expr)) tc:type-expr))
(define tc:make-variant-type-expr
  (lambda (sym lot)
    (tc:variant-type-expr sym lot)))

(deftype tc:make-variant-type-decl-type-expr
  (-> (symbol (list-of tc:type-expr)) tc:type-expr))
(define tc:make-variant-type-decl-type-expr
  ;; We keep this alias around for code in the old type checker
  (lambda (sym lot)
    (tc:make-variant-type-expr sym lot)))

(deftype tc:make-field-type-binding-type-expr
  (-> (symbol tc:type-expr) tc:type-expr))
(define tc:make-field-type-binding-type-expr
  (lambda (sym te)
    (tc:field-type-binding-type-expr sym te)))

(deftype tc:make-applied-type-expr
  (-> (tc:type-expr (list-of tc:type-expr)) tc:type-expr))
(define tc:make-applied-type-expr
  (lambda (te lot)
    (tc:applied-type-expr te lot)))


;;                 Type Testers


(deftype tc:variable-type-expr? (-> (tc:type-expr) boolean))
(define tc:variable-type-expr?
  (lambda (d)
    (cases tc:type-expr d
      (tc:variable-type-expr (lvar) #t)
      (else #f))))

(deftype tc:basic-type-expr? (-> (tc:type-expr) boolean))
(define tc:basic-type-expr?
  (lambda (d)
    (cases tc:type-expr d
      (tc:basic-type-expr (symbol) #t)
      (else #f))))

(deftype tc:non-variable-basic-type-expr? (-> (tc:type-expr) boolean))
(define tc:non-variable-basic-type-expr?
  ;; REQUIRES: d is not a type variable
  ;; ENSURES: result is true just when d is a basic type
  (lambda (d)
    (tc:basic-type-expr? d)))

(deftype tc:dots-type-expr? (-> (tc:type-expr) boolean))
(define tc:dots-type-expr?
  (lambda (d)
    (cases tc:type-expr d
      (tc:basic-type-expr (symbol) (eq? symbol '...))
      (else #f))))

(deftype tc:function-type-expr? (-> (tc:type-expr) boolean))
(define tc:function-type-expr?
  (lambda (d)
    (cases tc:type-expr d
      (tc:function-type-expr (arg-types result-type) #t)
      (else #f))))

(deftype tc:varargs-function-type-expr? (-> (tc:type-expr) boolean))
(define tc:varargs-function-type-expr?
  (lambda (d)
    (cases tc:type-expr d
      (tc:function-type-expr (arg-types result-type)
	 (and (not (null? arg-types))
	      (tc:dots-type-expr? (tc:last-item arg-types))))
      (else #f))))

(deftype tc:monotype? (-> (tc:type-expr) boolean))
(define tc:monotype?
  (lambda (te)
      (cases tc:type-expr te
	(tc:variable-type-expr (lvar) #f)
	(tc:basic-type-expr (symbol) #t)
	(tc:function-type-expr (arg-types result-type)
	  (and (tc:all tc:monotype? arg-types)
	       (tc:monotype? result-type)))
	(tc:intersection-type-expr (conjoined-types) #f)
	(tc:type-predicate-for-type-expr (type) (tc:monotype? type))
	(tc:variant-record-type-expr (variants)
				     (tc:all tc:monotype? variants))
	(tc:variant-type-expr (variant-name fields)
	    (tc:all tc:monotype? fields))
	(tc:field-type-binding-type-expr (field-name type) (tc:monotype? type))
     	(tc:applied-type-expr (operator-type operand-types)
	    (tc:all tc:monotype? operand-types))
	;; [[[ what about declaration and error, is this else enough?]]]
	(else #f)
	)))

(deftype tc:type-predicate-for-type-expr? (-> (tc:type-expr) boolean))
(define tc:type-predicate-for-type-expr?
  (lambda (d)
    (cases tc:type-expr d
      (tc:type-predicate-for-type-expr (type) #t)
      (else #f))))

(deftype tc:variant-record-type-expr? (-> (tc:type-expr) boolean))
(define tc:variant-record-type-expr?
  (lambda (d)
    (cases tc:type-expr d
      (tc:variant-record-type-expr (variants) #t)
      (else #f))))

(deftype tc:variant-type-expr? (-> (tc:type-expr) boolean))
(define tc:variant-type-expr?
  (lambda (d)
    (cases tc:type-expr d
      (tc:variant-type-expr (variant-name fields) #t)
      (else #f))))

(deftype tc:field-type-binding-type-expr? (-> (tc:type-expr) boolean))
(define tc:field-type-binding-type-expr?
  (lambda (d)
    (cases tc:type-expr d
      (tc:field-type-binding-type-expr (field-name type) #t)
      (else #f))))


(deftype tc:applied-type-expr? (-> (tc:type-expr) boolean))
(define tc:applied-type-expr?
  (lambda (d)
    (cases tc:type-expr d
      (tc:applied-type-expr (operator-type operand-types) #t)
      (else #f))))

(deftype tc:contains-function-type-expr? (-> (tc:type-expr) boolean))
(define tc:contains-function-type-expr?
  (lambda (type)
    (cases tc:type-expr type
	   (tc:function-type-expr (arg-types result-type) #t)
	   (tc:intersection-type-expr
	    (conjoined-types)
	    (not (null? (tc:filter tc:contains-function-type-expr? conjoined-types))))	   
	   (else #f))))


;;      Extraction Helpers
;; -------------------------------------------

(deftype tc:variable-type-expr->uniq-sym (-> (tc:type-expr) uniq-sym))
(define tc:variable-type-expr->uniq-sym
  (lambda (vt)
    ;; REQUIRES: vt is a variable-type-expr
    ;; EFFECT: returns a symbol formed by concatenating vt's name, \#, and
    ;;         vt's level
    (cases tc:type-expr vt
     (tc:variable-type-expr
       (lvar)
       (tc:logicalvar->uniq-sym lvar))
     (else
      (error "tc:variable-type-expr->uniq-sym: not a variable-type-expr: " vt)
      ))))


;;; [[[ check this ]]]
(deftype tc:uniq-sym->variable-type-expr (-> (uniq-sym) tc:type-expr))
(define tc:uniq-sym->variable-type-expr
  (letrec ((findit
	    (has-type
	     (-> (string number) number)
	     (lambda (s p)
	       (cond
		((= p (string-length s)) -1)
		((equal? (string-ref s p) #\#) p)
		(else (findit s (+ p 1))))))))
    (lambda (us)
      ;; EFFECT: returns a variable-type-expr with the appropriate name and
      ;;         level extracted from us
      (let* ((ustr (symbol->string us))
	     (pos (findit ustr 1)))	; don't check position 0, it can't
					; (legally) be there
	(if (not (= pos -1))
	    (tc:variable-type-expr
	     (tc:make-logicalvar 
	      ;(tc:symbol->logicalvar (string->symbol (substring ustr 0 pos)))
	      (string->symbol (substring ustr 0 pos))
	      (string->number (substring ustr (+ pos 1) (string-length ustr)))))
	    (error "tc:uniq-sym->variable-type-expr: invalid uniq-sym: "
		   us))))))


(deftype tc:basic-type-expr->symbol (-> (tc:type-expr) symbol))
(define tc:basic-type-expr->symbol
  (lambda (bt)
    (cases tc:type-expr bt
      (tc:basic-type-expr (symbol) symbol)
      (else
       (error "tc:basic-type-expr->symbol: not a basic-type-expr: "
	      bt)))))

(deftype tc:function-type-expr->arg-types
  (-> (tc:type-expr) (list-of tc:type-expr)))
(define tc:function-type-expr->arg-types
  (lambda (ft)
    (cases tc:type-expr ft
      (tc:function-type-expr (arg-types result-type) arg-types)
      (else
       (error "tc:function-type-expr->arg-types: not a function-type-expr: "
	      ft)))))

(deftype tc:function-type-expr->result-type (-> (tc:type-expr) tc:type-expr))
(define tc:function-type-expr->result-type
  (lambda (ft)
    (cases tc:type-expr ft
      (tc:function-type-expr (arg-types result-type) result-type)
      (else
       (error "tc:function-type-expr->result-type: not a function-type-expr: "
	      ft)))))

(deftype tc:intersection-type-expr->conjoined-types
  (-> (tc:type-expr) (list-of tc:type-expr)))
(define tc:intersection-type-expr->conjoined-types
  (lambda (it)
    (cases tc:type-expr it
      (tc:intersection-type-expr (conjoined-types) conjoined-types)
      (else
       (error
	"tc:intersection-type-expr->conjoined-types: not a intersection-type-expr: "
	it)))))

(deftype tc:type-predicate-for-type-expr->type
  (-> (tc:type-expr) tc:type-expr))
(define tc:type-predicate-for-type-expr->type
  (lambda (tpf)
    (cases tc:type-expr tpf
      (tc:type-predicate-for-type-expr (type) type)
      (else
       (error
	"tc:type-predicate-for-type-expr->type: not a type-predicate-for-type-expr: "
	tpf)))))

(deftype tc:variant-type-expr->variant-name
  (-> (tc:type-expr) symbol))
(define tc:variant-type-expr->variant-name
  (lambda (vt)
    (cases tc:type-expr vt
	   (tc:variant-type-expr (variant-name fields) variant-name)
	   (else
	    (error "tc:variant-type-expr->variant-name: not a variant-type-expr: "
		   vt)))))

(deftype tc:variant-type-expr->fields
  (-> (tc:type-expr) (list-of tc:type-expr)))
(define tc:variant-type-expr->fields
  (lambda (vt)
    (cases tc:type-expr vt
	  (tc:variant-type-expr (variant-name fields) fields)
	  (else
	   (error "tc:variant-type-expr->fields: not a variant-type-expr: "
		  vt)))))

(deftype tc:applied-type-expr->operator-type
  (-> (tc:type-expr) tc:type-expr))
(define tc:applied-type-expr->operator-type
  (lambda (at)
    (cases tc:type-expr at
      (tc:applied-type-expr (operator-type operand-types) operator-type)
      (else
       (error "tc:applied-type-expr->arg-type: not an applied-type-expr: "
	      at)))))

(deftype tc:applied-type-expr->operand-types
  (-> (tc:type-expr) (list-of tc:type-expr)))
(define tc:applied-type-expr->operand-types
  (lambda (at)
    (cases tc:type-expr at
      (tc:applied-type-expr (operator-type operand-types) operand-types)
      (else
       (error "tc:applied-type-expr->operand-types: not an applied-type-expr: "
	      at)))))

(deftype tc:variant-record-type-expr->variants
  (-> (tc:type-expr) (list-of tc:type-expr)))
(define tc:variant-record-type-expr->variants
  (lambda (vrt)
    (cases tc:type-expr vrt
	   (tc:variant-record-type-expr (variants) variants)
	   (else
	    (error "tc:variant-record-type-expr->variants: not a variant-record-type-expr: "
		   vrt)))))

(deftype tc:field-type-binding-type-expr->field-name
  (-> (tc:type-expr) symbol))
(define tc:field-type-binding-type-expr->field-name
  (lambda (ftb)
    (cases tc:type-expr ftb
	   (tc:field-type-binding-type-expr (field-name type) field-name)
	   (else
	    (error "tc:field-type-binding-type-expr->field-name: not a field-type-binding-type-expr: "
		   ftb)))))

(deftype tc:field-type-binding-type-expr->type
  (-> (tc:type-expr) tc:type-expr))
(define tc:field-type-binding-type-expr->type
  (lambda (ftb)
    (cases tc:type-expr ftb
	   (tc:field-type-binding-type-expr (field-name type) type)
	   (else
	    (error "tc:field-type-binding-type-expr->type: not a field-type-binding-type-expr: "
		   ftb)))))

(deftype tc:error-type-expr->error-record
  (-> (tc:type-expr) tc:error-record))
(define tc:error-type-expr->error-record
  (lambda (e)
    (cases tc:type-expr e
	   (tc:error-type-expr (error-record) error-record)
	   (else
	    (error "tc:error-type-expr->error-record: not a error-type-expr: "
		   e)))))

;;                 OTHER HELPERS
;; ------------------------------------------

(deftype tc:basic-type-expr-eq? (-> (tc:type-expr tc:type-expr) boolean))
(define tc:basic-type-expr-eq?
  (lambda (typ1 typ2)
    ;; REQUIRE: typ1 and typ2 are basic-type-exprs
    ;; EFFECT: returns true if typ1 and typ2 are equivalent
    (eq? (tc:basic-type-expr->symbol typ1)
	 (tc:basic-type-expr->symbol typ2))))

(deftype tc:uniq-sym->symbol (-> (uniq-sym) symbol))
(define tc:uniq-sym->symbol
  (lambda (us) us))

(deftype tc:type-expr-instance
  (-> (tc:type-expr) tc:type-expr))
(define tc:type-expr-instance
  (lambda (te)
    ;; ENSURES: result is a tc:type-expr that is
    ;;          like te, but in which the type variables are
    ;;          instantiated using unique variable-type-expr
    (let ((env
	   (has-type
	    (tc:env-of symbol tc:type-expr)
	    (tc:env-empty))))
      (letrec
	  ((instancer
	    (has-type
	     (-> (tc:type-expr) tc:type-expr)
	     (lambda (te)
	       (cases tc:type-expr te
		 (tc:variable-type-expr (lvar)
		  (let ((us (tc:variable-type-expr->uniq-sym te)))
		    (if (tc:env-bound? us env)
			(tc:env-value us env)
			(let ((var (tc:new-variable-type-expr)))
			  (set! env (tc:env-update us var env))
			  var))))
		 (tc:basic-type-expr (symbol) te)
		 (tc:function-type-expr (arg-types result-type)
		   (tc:function-type-expr
		    (map instancer arg-types)
		    (instancer result-type)))
		 (tc:intersection-type-expr (conjoined-types)
		   (tc:intersection-type-expr
		    (map instancer conjoined-types)))
		 (tc:type-predicate-for-type-expr (type)
		   (tc:type-predicate-for-type-expr (instancer type)))
		 (tc:variant-record-type-expr (variants)
		   (tc:variant-record-type-expr
		    (map instancer variants)))
		 (tc:variant-type-expr
		  (variant-name fields)
		  (tc:variant-type-expr
		   variant-name
		   (map instancer fields)))
		 (tc:field-type-binding-type-expr
		  (field-name type)
		  (tc:field-type-binding-type-expr
		   field-name
		   (instancer type)))
		 (tc:applied-type-expr (operator-type operand-types)
		   (tc:applied-type-expr
		    (instancer operator-type)
		    (map instancer operand-types)))
		 (else te) ;; for delcarations and error
		 )))))	   
	(instancer te)))))

(deftype tc:variable-type-expr-new-level
  (-> (tc:type-expr number) tc:type-expr))
(define tc:variable-type-expr-new-level
  (lambda (t n)
    ;; REQUIRES: t is a variable-type-expr
    ;; EFFECT: returns a copy of t with the level field set to n
    (cases tc:type-expr t
      (tc:variable-type-expr
       (lvar)
       (tc:variable-type-expr (tc:make-logicalvar
			       (tc:logicalvar->name lvar)
			       n)))
      (else
       (error "tc:variable-type-expr-new-level: not a variable-type-expr: " t)
       ))))

;; see the has-type-trusted below for the type
(define tc:type-expr-uniquify
  (let ((*tc:unique-var-number* 0))
    (letrec ((uniquify
	      (has-type
	       (-> (tc:type-expr) tc:type-expr)
	       (lambda (te)
		 (cases tc:type-expr te
		   (tc:variable-type-expr (lvar)
		     (tc:variable-type-expr-new-level
		      te *tc:unique-var-number*))
		   (tc:basic-type-expr (symbol) te)
		   (tc:function-type-expr (arg-types result-type)
		     (tc:function-type-expr
		      (map uniquify arg-types)
		      (uniquify result-type)))
		   (tc:intersection-type-expr (conjoined-types)
		     (tc:intersection-type-expr
		      (map uniquify conjoined-types)))
		   (tc:type-predicate-for-type-expr (type)
		     (tc:type-predicate-for-type-expr
		      (uniquify type)))
		   (tc:variant-record-type-expr
		    (variants)
		    (tc:variant-record-type-expr
		     (map uniquify variants)))
		   (tc:variant-type-expr
		    (variant-name fields)
		    (tc:variant-type-expr
		     variant-name
		     (map uniquify fields)))
		   (tc:field-type-binding-type-expr
		    (field-name type)
		    (tc:field-type-binding-type-expr
		     field-name (uniquify type)))
		   (tc:applied-type-expr (operator-type operand-types)
		     (tc:applied-type-expr
		      (uniquify operator-type)
		      (map uniquify operand-types)))
		   (tc:declaration-type-expr (pi) te)    ;;; [[[FIXME:  I don't think this is this easy]]]
		   (tc:error-type-expr (error-record) te)
		   
		   )))))
      (has-type-trusted
       (all-of (-> (tc:type-expr) tc:type-expr)
	       (-> ((pair-of tc:type-expr tc:type-expr))
		   (pair-of tc:type-expr tc:type-expr)))
       (lambda (type-expr)
	 ;; ENSURES: result is a tc:type-expr, or pair of tc:type-expr, that is
	 ;;          like type-expr, but in which the type variables are
	 ;;          unique (in this execution)
	 ;; NOTE: This procedure must handle pairs of tc:type-expr so that
	 ;;       the binding of variables between the abstract and
	 ;;       representation types in a defrep can be maintained.
	 (set! *tc:unique-var-number* (+ *tc:unique-var-number* 1))
	 (if (not (tc:type-expr? type-expr)) ; implies (pair? type-expr)
	     (cons (uniquify (car type-expr))
		   (uniquify (cdr type-expr)))
	     (uniquify type-expr)))))))

;; For dealing with dots types
(deftype tc:fix-length (-> (tc:type-expr tc:type-expr) tc:type-expr))
(define tc:fix-length
  (letrec ((get-dot-type
	    (has-type
	     (-> ((list-of tc:type-expr)) tc:type-expr)
	     (lambda (ls)
	       ;; REQUIRES: ls at least one type followed by '...' in it
	       ;; EFFECT: returns the element of the ls before the '...'
	       (if (tc:dots-type-expr? (cadr ls))
		   (car ls)
		   (get-dot-type (cdr ls))))))
	   (preceed-dot-type
	    (has-type
	     (-> ((list-of tc:type-expr)) (list-of tc:type-expr))
	     (lambda (ls)
	       ;; REQUIRES: ls at least one type followed by '...' in it
	       ;; EFFECT: returns the part of the list before the '...'
	       ;;         not including the type just before the '...'
	       (cond
		((null? (cddr ls)) '())
		((tc:dots-type-expr? (caddr ls)) (list (car ls)))
		(else (cons (car ls) (preceed-dot-type (cdr ls))))))))
	   (enough-copies
	    (has-type
	     (-> (number tc:type-expr) (list-of tc:type-expr))
	     (lambda (n dotted)
	       (if (<= n 0)
		   '()
		   (cons dotted (enough-copies (- n 1) dotted)))))))
    (lambda (dot-type arrow-type)
      ;; REQUIRES: dot-type is an arrow type with '...' in it
      ;; EFFECT: result is a type like dot-type but with an appropriate
      ;;         number of copies of types to fill out the ... in hopes of
      ;;         matching arrow-type
      (let ((type-list (tc:function-type-expr->arg-types dot-type))
	    (arg-list (tc:function-type-expr->arg-types arrow-type)))
	(tc:make-function-type-expr
	 (let* ((dot-type (get-dot-type type-list))
		(before (preceed-dot-type type-list)))
	   (append
	    before
	    (enough-copies (- (length arg-list) (length before))
			   dot-type)))
	 (tc:function-type-expr->result-type dot-type))))))

(deftype tc:type-expr-replace
  (-> (tc:type-expr tc:type-expr tc:type-expr) tc:type-expr))
(define tc:type-expr-replace
  (lambda (old new te)
    (if (equal? te old)
	new
	;; Otherwise we recurse into te
	(let ((replace (lambda (x) (tc:type-expr-replace old new x))))
	  (cases tc:type-expr te
		 (tc:function-type-expr
		  (arg-types result-type)
		  (tc:function-type-expr (map replace arg-types)
					 (replace result-type)))
		 (tc:intersection-type-expr
		  (conjoined-types)
		  (tc:intersection-type-expr (map replace conjoined-types)))
		 (tc:type-predicate-for-type-expr
		  (type)
		  (tc:type-predicate-for-type-expr (replace type)))
		 (tc:variant-record-type-expr
		  (variants)
		  (tc:variant-record-type-expr (map replace variants)))
		 (tc:applied-type-expr
		  (operator-type operand-types)
		  (tc:applied-type-expr (replace operator-type)
					(map replace operand-types)))
		 (tc:variant-type-expr
		  (variant-name fields)
		  (tc:variant-type-expr variant-name
					(map replace fields)))
		 (tc:field-type-binding-type-expr
		  (field-name type)
		  (tc:field-type-binding-type-expr field-name
						   (replace type)))
		 (tc:declaration-type-expr
		  (pi)
		  (tc:declaration-type-expr
		   (tc:env-map (lambda (k v)
				 (tc:bind-type-expr
				  (tc:binding->binder v)
				  (replace (tc:binding->type-expr v))))
			       pi)))
		 ;; Just stop.
		 (else te))))))


;;;                 ERROR RECORDS                     ;;;
;;; This is the former contents of tc-error.scm

(define-datatype tc:error-record tc:error-record?
  (tc:mismatch-error-record
   (syn datum?) (expected tc:type-expr?) (inferred tc:type-expr?))
  (tc:composite-error-record
   (errors (list-of tc:error-record?)))
  (tc:if-subtype-error-record
   (left-exp tc:scheme-exp?) (left-type tc:type-expr?)
   (right-exp tc:scheme-exp?) (right-type tc:type-expr?))
  (tc:apply-non-proc-error-record
   (exp tc:scheme-exp?)
   (type tc:type-expr?))
  (tc:op-arg-mismatch-error-record
   (oper-exp tc:scheme-exp?)
   (args-list (list-of tc:scheme-exp?))
   (operator-type tc:type-expr?)
   (args-types (list-of tc:type-expr?)))
  (tc:varref-error-record
   (syn tc:scheme-exp?))
  (tc:badtest-error-record
   (test-record tc:error-record?))
  (tc:defined-lambda-mismatch-error-record
   (syn tc:scheme-exp?) (declared-type tc:type-expr?))
  (tc:named-lambda-subtyping-error-record
   (syn tc:scheme-exp?)
   (name symbol?)
   (declared-type tc:type-expr?)
   ;; [[[FIXME: can't seem to get proper inferred type in here...]]]
   )
  (tc:duplicate-let-binding-error-record
   (syn tc:scheme-exp?)
   (dups (list-of symbol?)))
  (tc:cases-clause-bad-fields-error-record
   (syn tc:scheme-cases-clause?)
   (fields (list-of symbol?))
   (field-types (list-of tc:type-expr?)))
  (tc:cases-on-non-variant-error-record
   (syn tc:scheme-exp?)
   (sym symbol?))
  (tc:cases-with-bad-clause-usage-error-record
   (syn tc:scheme-exp?)
   (variant-type tc:type-expr?))
  (tc:duplicate-formals-error-record
   (syn tc:scheme-formals?)
   (dups (list-of symbol?)))
  (tc:non-void-body-error-record
   (syn tc:scheme-exp?)
   (inferred tc:type-expr?))
  ;;[[IMPROVE: it would be nice to have the second type on this...]]
  (tc:set!-mismatch-error-record
   (variable symbol?)
   (exp tc:scheme-exp?)
   (decl-type tc:type-expr?)
   (inf-type tc:type-expr?))
  (tc:cond-clauses-mismatch-error-record
   (exp tc:scheme-exp?)
   (clause-types (list-of tc:type-expr?)))
  (tc:cases-clauses-mismatch-error-record
   (exp tc:scheme-exp?)
   (clause-types (list-of tc:type-expr?)))
  (tc:has-type-mismatch-error-record
   (exp tc:scheme-exp?)
   (decl-type tc:type-expr?)
   (inf-type tc:type-expr?))
  (tc:forall-bad-body-error-record
   (exp tc:scheme-exp?)
   (inf-type tc:type-expr?)
   (exp-type tc:type-expr?))
  (tc:define-type-mismatch-error-record
   (variable symbol?)
   (exp tc:scheme-definition?)
   (type tc:type-expr?))
  (tc:define-datatype-name-conflict-error-record
   (syn tc:scheme-definition?)
   (type-name symbol?))
  (tc:unknown-module-error-record
   (syn tc:scheme-module-name?)
   (mod-identifier symbol?))
  (tc:bad-file-error-record
   (pos tc:position?)
   (path string?))
  (tc:deftype-redeclaration-error-record
   (syn tc:scheme-definition?)
   (identifier symbol?))
  (tc:module-missing-deftype-error-record
   (syn tc:scheme-module-name?)
   (mod-name symbol?)
   (path string?))
  ;; for testing
  (tc:simple-error-record
   (msg string?))
  )

(deftype tc:mismatch-error-maker
  (-> (datum tc:type-expr tc:type-expr) tc:type-expr))
(define tc:mismatch-error-maker
  (lambda (syn expected inferred)
    (if (> (type-tracing-level) 0)
	(begin
	  (displayln "ERROR: tc:mismatch-error-maker")
	  (displayln "expected: " expected)
	  (displayln "inferred: " inferred)))
    (if (tc:error-type-expr? inferred)
	inferred
	(tc:error-type-expr
	 (tc:mismatch-error-record syn expected inferred)))))

(deftype tc:rule-error-maker
  (-> (datum (list-of (tc:tree-of tc:type-expr)))
      tc:type-expr))
(define tc:rule-error-maker
  (lambda (syn trees)
    (has-type-trusted
      tc:type-expr
      (let ((rule-error-maker
	     (cond
	      ;; ((tc:scheme-form? syn) tc:scheme-form-error-maker)
	      ((tc:scheme-exp? syn) tc:scheme-exp-error-maker)
	      ;; ((tc:scheme-datum? syn) tc:scheme-datum-error-maker)
	      ((tc:scheme-formals? syn) tc:scheme-formals-error-maker) ;;[[DONE]]
	      ((tc:scheme-definition? syn) tc:scheme-definition-error-maker)
	      ;; ((tc:variant-decl? syn) tc:variant-decl-record-error-maker)
	      ;; ((tc:field-pred-binding? syn) tc:field-pred-binding-record-error-maker)
	      ((tc:scheme-cond-clause? syn) tc:scheme-cond-clause-error-maker)
	      ;; ((tc:scheme-case-clause? syn) tc:scheme-case-clause-error-maker)
	      ;; ((tc:scheme-binding-spec? syn) tc:scheme-binding-spec-error-maker)
	      ;; ((tc:scheme-cases-clause? syn) tc:scheme-cases-clause-error-maker)
	      ;; ((tc:scheme-iteration-spec? syn) tc:scheme-iteration-spec-error-maker)
	      ;; ((tc:scheme-program? syn) tc:scheme-program-error-maker)
	      ;; ((tc:input-type-exp? syn) tc:input-type-exp-error-maker)
	      ((tc:scheme-module-name? syn) tc:scheme-module-name-error-maker)

	      ;; Cases handled by the defaults
	      ;; tc:scheme-body
	    
	      ;; more cases go here	    
	      (else tc:generic-error-maker)
	      )))
	(rule-error-maker syn trees)))))


(deftype tc:scheme-module-name-error-maker
  (-> (tc:scheme-module-name (list-of (tc:tree-of tc:type-expr)))
      tc:type-expr))
(define tc:scheme-module-name-error-maker
  (lambda (syn trees)
    (let ((tree-tops (map tc:root trees)))
      (if (tc:contains-errors? tree-tops)
	  (tc:composite-error-maker tree-tops)
	  ;; Error missing deftype for provides list
	  (tc:error-type-expr
	   (let ((path (tc:scheme-module-name->file-location syn)))
	     (tc:module-missing-deftype-error-record
	      syn
	      (tc:extract-name path)
	      path)))))))

(deftype tc:scheme-definition-error-maker
  (-> (tc:scheme-definition (list-of (tc:tree-of tc:type-expr)))
      tc:type-expr))
(define tc:scheme-definition-error-maker
  (lambda (syn trees)
    (let ((tree-tops (map tc:root trees)))
      (cases tc:scheme-definition syn
	     (tc:simple-define
	      (position variable exp)
	      (if (tc:contains-errors? tree-tops)
		  (tc:composite-error-maker tree-tops)
		  ;; Error on subtyping part
		  (tc:error-type-expr
		   (tc:define-type-mismatch-error-record
		    variable syn (car tree-tops)))))
	     (tc:define-datatype
	      (position type-name type-predicate-name variants)
	      (if (tc:contains-errors? tree-tops)
		  (tc:composite-error-maker tree-tops)
		  ;; Error with the side condition--type-name already exists
		  (tc:error-type-expr
		   (tc:define-datatype-name-conflict-error-record syn type-name))))
	     (tc:deftype
	      (position variable type)
	      (if (tc:contains-errors? tree-tops)
		  (tc:composite-error-maker tree-tops)
		  ;; Error!  Redeclaration
		  (tc:error-type-expr
		   (tc:deftype-redeclaration-error-record syn variable))))
	     (else (tc:generic-error-maker syn trees))))))

(deftype tc:scheme-formals-error-maker
  (-> (tc:scheme-formals (list-of (tc:tree-of tc:type-expr))) tc:type-expr))
(define tc:scheme-formals-error-maker
  (lambda (fs trees)
    (tc:error-type-expr
     (tc:duplicate-formals-error-record
      fs
      (tc:find-duplicates (tc:scheme-formals->names fs))))))
      
(deftype tc:scheme-exp-error-maker
  (-> (tc:scheme-exp (list-of (tc:tree-of tc:type-expr))) tc:type-expr))
(define tc:scheme-exp-error-maker
  (lambda (syn trees)
    (let ((tree-tops (map tc:root trees)))
      (cases tc:scheme-exp syn
	(tc:procedure-call
	 (position operator operands)
	 (cond
	  ((tc:contains-errors? tree-tops)
	   (tc:composite-error-maker tree-tops))
	  ;; Application of a non-procedure
	  ((not (tc:contains-function-type-expr? (car tree-tops)))
	   (tc:error-type-expr
	    (tc:apply-non-proc-error-record
	     operator (car tree-tops))))
	  ;; Op and arg types don't match (subtype error)
	  (else 		  
	   (tc:error-type-expr
	    (tc:op-arg-mismatch-error-record
	     operator operands
	     (car tree-tops)
	     (cdr tree-tops))
	    ))))
	(tc:conditional-exp
	 (position test consequent alternate)
	 (let ((tests-type (car tree-tops)))
	   (cond
	    ((and (tc:error-type-expr? tests-type)		       
		  (tc:mismatch-error-record?
		   (tc:error-type-expr->error-record tests-type)))
	     (tc:error-type-expr
	      (tc:badtest-error-record
	       (tc:error-type-expr->error-record tests-type))))
	    ((tc:contains-errors? tree-tops)
	     (tc:composite-error-maker tree-tops))
	    (else
	     (tc:error-type-expr
	      (tc:if-subtype-error-record
	       consequent (cadr tree-tops)
	       alternate (caddr tree-tops)))))))
	(tc:conditional-stmt
	 (position test consequent)
	 (let ((tests-type (car tree-tops)))
	   (cond
	    ((and (tc:error-type-expr? tests-type)
		  (tc:mismatch-error-record?
		   (tc:error-type-expr->error-record tests-type)))
	     (tc:error-type-expr
	      (tc:badtest-error-record
	       (tc:error-type-expr->error-record tests-type))))
	    ((null? (cdr tree-tops))
	     (tc:composite-error-maker tree-tops))
	    (else
	     (let ((conseq-type (cadr tree-tops)))
	       (if (and (tc:error-type-expr? conseq-type)
			(tc:mismatch-error-record?
			 (tc:error-type-expr->error-record
			  conseq-type)))
		   (tc:error-type-expr
		    (tc:non-void-body-error-record
		     consequent
		     (tc:mismatch-error-record->inferred-type
		      (tc:error-type-expr->error-record
		       conseq-type))))
		   (tc:composite-error-maker tree-tops)))))))
	(tc:assignment
	 (position variable exp)
	 (if (tc:contains-errors? tree-tops)
	     (tc:composite-error-maker tree-tops)
	     ;; Error on subtyping part
	     (tc:error-type-expr
	      (tc:set!-mismatch-error-record variable exp (car tree-tops)
					     (cadr tree-tops)))))
	(tc:cond-exp
	 (position clauses)
	 (if (tc:contains-errors? tree-tops)
	     (tc:composite-error-maker tree-tops)
	     ;; Error on subtyping part
	     (tc:error-type-expr
	      (tc:cond-clauses-mismatch-error-record syn tree-tops))))
	(tc:cases-exp
	 (position type-name exp clauses)
	 (if (tc:contains-errors? tree-tops)
	     (tc:composite-error-maker tree-tops)
	     ;; Error on subtyping part
	     (tc:error-type-expr
	      (tc:cases-clauses-mismatch-error-record
	       syn (cdr tree-tops)))))
	(tc:cases-stmt
	 (position type-name exp clauses)
	 (if (tc:contains-errors? tree-tops)
	     (tc:composite-error-maker tree-tops)
	     ;; Error on subtyping part
	     (tc:error-type-expr
	      (tc:cases-clauses-mismatch-error-record
	       syn (cdr tree-tops)))))
	(tc:has-type-exp
	 (position type exp)
	 (if (tc:contains-errors? tree-tops)
	     (tc:composite-error-maker tree-tops)
	     ;; Error on subtyping part
	     (tc:error-type-expr
	      (tc:has-type-mismatch-error-record
	       exp (cadr tree-tops) (car tree-tops)))))
	(tc:forall-exp
	 (position formals exp)
	 (let ((body-type (cadr tree-tops)))
	   (if (and (tc:error-type-expr? body-type)
		    (tc:mismatch-error-record?
		     (tc:error-type-expr->error-record body-type)))
	       (let ((body-error-record
		      (tc:error-type-expr->error-record body-type)))
		 (tc:error-type-expr
		  (tc:forall-bad-body-error-record
		   exp
		   (tc:mismatch-error-record->inferred-type
		    body-error-record)
		   (tc:mismatch-error-record->expected-type
		    body-error-record))))
	       (tc:composite-error-maker tree-tops))))
	(tc:varref
	 (position variable)
	 (tc:error-type-expr (tc:varref-error-record syn)))

	;; Errors that could maybe be nicer...
	;;cond-stmt
	;;case-exp
	;;case-stmt
	;;do-exp
	;;do-stmt
	;;define-datatype-exp
	;;constructor-for-variant-type-exp
	     
	(else
	 (tc:generic-error-maker syn trees))))))

(deftype tc:scheme-cond-clause-error-maker
  (-> (tc:scheme-cond-clause (list-of (tc:tree-of tc:type-expr)))
      tc:type-expr))
(define tc:scheme-cond-clause-error-maker
  (lambda (syn trees)
    (let ((tree-tops (map tc:root trees)))
      (if (or (tc:normal-cond-clause? syn)
	      (tc:test=>cond-clause? syn))
	  (let ((tests-type (car tree-tops)))
	    (if (and (tc:error-type-expr? tests-type)
		     (tc:mismatch-error-record?
		      (tc:error-type-expr->error-record tests-type)))
		(tc:error-type-expr
		 (tc:badtest-error-record
		  (tc:error-type-expr->error-record tests-type)))
		(tc:composite-error-maker tree-tops)))	  
	  (tc:composite-error-maker tree-tops)))))

(deftype tc:generic-error-maker
  (forall (syn t)
    (-> (syn (list-of (tc:tree-of t))) tc:type-expr)))
(define tc:generic-error-maker
  (lambda (syn trees)
    (let ((tree-tops (map tc:root trees)))
      ;; (displayln "GENERIC CALLED: " syn)
      ;; (displayln "TREE-TOPS: " tree-tops)
      (if (tc:contains-errors? tree-tops)
	  (tc:composite-error-maker tree-tops)
	  (tc:error-type-expr (tc:simple-error-record "GENERIC RULE ERROR"))))))

(deftype tc:composite-error-maker
    (-> ((list-of tc:type-expr)) tc:type-expr))
(define tc:composite-error-maker
  (lambda (tree-tops)
    (tc:error-type-expr
     (tc:composite-error-record
      (tc:extract-errors tree-tops)))))

(deftype tc:varref-error-maker
  (-> (tc:scheme-exp) tc:type-expr))
(define tc:varref-error-maker
  (lambda (syn)
    (tc:error-type-expr (tc:varref-error-record syn))))

;;[[[FIXME: elminate this with rule if in annotate]]]
(deftype tc:unknown-module-error-maker
  (-> (tc:scheme-module-name symbol) tc:type-expr))
(define tc:unknown-module-error-maker
  (lambda (mod-name sym)
    (tc:error-type-expr (tc:unknown-module-error-record mod-name sym))))

(deftype tc:defined-lambda-mismatch-error-maker
  (-> (tc:scheme-exp tc:type-expr) tc:type-expr))
(define tc:defined-lambda-mismatch-error-maker
  (lambda (syn declared-type)
    (tc:error-type-expr
     (tc:defined-lambda-mismatch-error-record syn declared-type))))

(deftype tc:named-lambda-subtyping-error-maker
  (-> (tc:scheme-exp symbol tc:type-expr) tc:type-expr))
(define tc:named-lambda-subtyping-error-maker
  (lambda (syn name declared-type)
    (tc:error-type-expr
     (tc:named-lambda-subtyping-error-record
      syn name declared-type))))

(deftype tc:duplicate-let-binding-error-maker
  (-> (tc:scheme-exp (list-of symbol)) tc:type-expr))
(define tc:duplicate-let-binding-error-maker
  (lambda (syn dups)
    (tc:error-type-expr (tc:duplicate-let-binding-error-record syn dups))))

(deftype tc:cases-clause-bad-fields-error-maker
  (-> (tc:scheme-cases-clause (list-of symbol) (list-of tc:type-expr))
      tc:type-expr))
(define tc:cases-clause-bad-fields-error-maker
  (lambda (syn fields field-types)
    (tc:error-type-expr
     (tc:cases-clause-bad-fields-error-record syn fields field-types))))

(deftype tc:cases-on-non-variant-record-error-maker
  (-> (tc:scheme-exp symbol) tc:type-expr))
(define tc:cases-on-non-variant-record-error-maker
  (lambda (syn sym)
    (tc:error-type-expr (tc:cases-on-non-variant-error-record syn sym))))

(deftype tc:cases-with-bad-clause-usage-error-maker
  (-> (tc:scheme-exp tc:type-expr) tc:type-expr))
(define tc:cases-with-bad-clause-usage-error-maker
  (lambda (syn variant-type)
    (tc:error-type-expr (tc:cases-with-bad-clause-usage-error-record syn variant-type))))

(deftype tc:bad-file-error-maker
  (-> (tc:position string) tc:type-expr))
(define tc:bad-file-error-maker
  (lambda (pos path)
    (tc:error-type-expr (tc:bad-file-error-record pos path))))

;; for example, if we have an error on syn for if statement:
;; (cond
;;   ((is-error? (tc:root (car trees)))
;;      (make message for bad test expression, use string inside the first tree))
;;   ((contains-error? trees)
;;      (there was a type error in a sub expression, we should use it as ours))
;;   (else
;;      (this error was caused by failure in subtyping judgement)))
     
(deftype tc:simple-error-maker (-> (string) tc:type-expr))
(define tc:simple-error-maker
  (lambda (msg)
    (tc:error-type-expr
     (tc:simple-error-record msg))))

(deftype tc:mismatch-error-record?
  (-> (tc:error-record) boolean))
(define tc:mismatch-error-record?
  (lambda (er)
    (cases tc:error-record er
	   (tc:mismatch-error-record (syn expected inferred) #t)
	   (else #f))))

(deftype tc:mismatch-error-record->inferred-type
  (-> (tc:error-record) tc:type-expr))
(define tc:mismatch-error-record->inferred-type
  (lambda (mer)
    (cases tc:error-record mer
	   (tc:mismatch-error-record (syn expected inferred) inferred)
	   (else (error "tc:mismatch-error-record->inferred called on non bad record: " mer)))))

(deftype tc:mismatch-error-record->expected-type
  (-> (tc:error-record) tc:type-expr))
(define tc:mismatch-error-record->expected-type
  (lambda (mer)
    (cases tc:error-record mer
	   (tc:mismatch-error-record (syn expected inferred) expected)
	   (else (error "tc:mismatch-error-record->expected called on non bad record: " mer)))))

(deftype tc:composite-error-record?
  (-> (tc:error-record) boolean))
(define tc:composite-error-record?
  (lambda (er)
    (cases tc:error-record er
	   (tc:composite-error-record (errors) #t)
	   (else #f))))

(deftype tc:contains-errors?
  (-> ((list-of tc:type-expr)) boolean))
(define tc:contains-errors?
  (lambda (types)
    (not (null? (tc:filter tc:error-type-expr? types)))))

(deftype tc:extract-errors
  (-> ((list-of tc:type-expr)) (list-of tc:error-record)))
(define tc:extract-errors
  (lambda (types)
    (map tc:error-type-expr->error-record (tc:filter tc:error-type-expr? types))))	


;;;         TYPE BINDINGS                              ;;;
;;; Type bindings are pairs whose car is a symbol from the
;;; set {let letrec lambda} and whose cdr is a type expression.

(deftype tc:binding? (type-predicate-for tc:binding))
(define tc:binding?
  (lambda (d)
    (and (pair? d)
	 (or (eq? 'let (car d))
	     (eq? 'letrec (car d))
	     (eq? 'lambda (car d)))
	 (tc:type-expr? (cdr d)))))
  
(deftype tc:bind-type-expr (-> (symbol tc:type-expr) tc:binding))
(define tc:bind-type-expr
  (lambda (s te)
    (make-pair (has-type symbol s) (has-type tc:type-expr te))))

(deftype tc:bound? (-> (symbol tc:binding) boolean))
(define tc:bound?
  (lambda (s b)
    (eq? (pair-first (has-type (pair-of symbol tc:type-expr) b)) s)))

(deftype tc:binding->type-expr (-> (tc:binding) tc:type-expr))
(define tc:binding->type-expr
  (lambda (b)
    (pair-second b)))

(deftype tc:binding->binder (-> (tc:binding) symbol))
(define tc:binding->binder
  (lambda (b) (pair-first b)))

) ;; end module
